!===============================================================================
!
! bqcdi.F90 - main program and read/write of parameters
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 1998-2010 Hinnerk Stueben, Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
program bqcd
  use module_cg
  use typedef_flags
  use typedef_para
  use module_input
  use module_function_decl
  use module_switches
  use module_lattice
  implicit none
  
  type(type_para)                       :: para
  type(hmc_conf), dimension(MAX_TEMPER) :: conf
  type(type_flags)                      :: flags
  SECONDS                               :: time0, sekunden

#ifdef QPACE
  call init_cellddsolver()
#endif

#ifdef QMP
  call init_qmp(0)
#else
  call comm_init()
#endif

  call place_process()
  !!call comm_finalize()
  !!stop

  time0 = sekunden()  ! start/initialize timer

  TIMING_START(timing_bin_total)


  call get_flags(flags)
  call begin(UREC, "Job")
  call input_read(flags%input)
#ifdef IBM
  call bgl_process_mapping()
#endif
  call init_para(para, flags)

#ifdef QMP
  call init_qmp_node(para%NPE(1),para%NPE(2),para%NPE(3),para%NPE(4))
#endif

  call init_counter(para, flags)
  call init_ran(para, flags)
  call init_cooling(input%measure_cooling_list)

  call set_fmt_ensemble(para%n_temper)
  call check_fmt(para%run, para%n_temper, para%maxtraj, para%L(4) - 1)

  call init_common(para)
  call init_modules()
  call init_d()
  call init_dd()

  call write_header(para, flags)

!!call remez()
!!stop
  call init_bc()
  call init_cg_para()
  call init_cg_stat_all()
  call init_xbound()
#ifndef D500
  call init_xbound_r4()
#endif
#ifdef BAGEL
  call init_wfm(NX,NY,NZ,NT,NPE) !!! for bagel
#endif

  call init_confs2(para, conf)

  call check_former(para%n_temper, conf)

!!  call ieo2local_test()
!!  call coord_test()
!!  call xbound_test()
!!  call comm_basic_test()  ! MPI only

  call mc(para, conf)

  if (.not. switches%measurement_only) then
     call conf_write(.true., para, conf)
     call write_counter(para%maxtraj)
     call write_ran()
     call write_plaq()
  endif

  TIMING_STOP(timing_bin_total)

  call write_footer(time0)
  call end(UREC, "Job")

#ifdef BAGEL
  call end_wfm()  !!! for bagel
#endif
  call comm_finalize()
end

!-------------------------------------------------------------------------------
subroutine init_para(para, flags)

  ! initialises module_para, module_switches and module_mre

  use typedef_flags
  use typedef_para
  use module_bqcd
  use module_input
  use module_mre
  use module_mre2
  use module_switches
  use module_switches
  use module_action
  use module_evec
  use module_dd
  implicit none

  type(type_para)  :: para
  type(type_flags) :: flags
  integer          :: i
  logical          :: quenched, dynamical, clover, h_ext

  quenched   = .false.
  dynamical  = .false.
  clover     = .false.
  h_ext      = .false.

  ddlat           = input%ddlattice

  para%run         = input%run
  para%L           = input%lattice
  para%NPE         = input%processes
  para%process_mapping = input%process_mapping
  para%bc_fermions = input%boundary_conditions_fermions
  para%gamma_index = input%gamma_index
  para%n_temper    = input%ensembles
  para%nstd        = input%tempering_steps_without
  para%nforce      = input%hmc_accept_first
  para%ntraj       = input%mc_steps
  para%maxtraj     = input%mc_total_steps
  para%nsave       = input%mc_save_frequency
  para%c_cg_rest   = input%solver_rest
  para%cg_maxiter  = input%solver_maxiter
  para%cg_log      = input%solver_ignore_no_convergence
  para%cg_check    = input%solver_check_solution
  mre_n_vec        = input%solver_mre_vectors
  mre2_n_vec        = input%solver_mre_vectors

  switches%measure_rhmc_forces = (input%measure_rhmc_forces /= 0)
  switches%measure_minmax      = (input%measure_minmax /= 0)
  switches%measure_correlations= (input%measure_correlations /= 0)
  neig = input%measure_minmax

  gauge_action= input%gauge_action
  fermi_action= input%fermi_action

  call check_bc_fermions(para%bc_fermions, para%gamma_index)

  read(para%c_cg_rest   , *) para%cg_rest

 if (para%n_temper <= 0) call die("init_para(): n_temper <= 0")
 if (para%n_temper > MAX_TEMPER) call die("init_para(): n_temper > max_temper")
  
  do i = 1, para%n_temper
     para%c_hmc(i)%beta        = input%beta(i)
     para%c_hmc(i)%kappa       = input%kappa(i)
     para%c_hmc(i)%csw         = input%csw(i)
     para%c_hmc(i)%h           = input%h(i)
     para%c_hmc(i)%traj_length = input%hmc_trajectory_length(i)
     para%c_hmc(i)%ntau        = input%hmc_steps(i)
     para%c_hmc(i)%rho         = input%hmc_rho(i)
     para%c_hmc(i)%m_scale     = input%hmc_m_scale(i)
     para%c_hmc(i)%m_scale2    = input%hmc_m_scale2(i)
     para%info_file(i)         = input%start_info_file(i)

     read(para%c_hmc(i)%beta,       *) para%hmc(i)%beta
     read(para%c_hmc(i)%kappa,      *) para%hmc(i)%kappa
     read(para%c_hmc(i)%csw,        *) para%hmc(i)%csw
     read(para%c_hmc(i)%h,          *) para%hmc(i)%h
     read(para%c_hmc(i)%traj_length,*) para%hmc(i)%traj_length
     read(para%c_hmc(i)%ntau,       *) para%hmc(i)%ntau
     read(para%c_hmc(i)%rho,        *) para%hmc(i)%rho
     read(para%c_hmc(i)%m_scale,    *) para%hmc(i)%m_scale
     read(para%c_hmc(i)%m_scale2,   *) para%hmc(i)%m_scale2

     para%c_hmc(i)%kappa_strange   = input%kappa_strange(i)
     para%c_hmc(i)%rho2            = input%hmc_rho2(i)
     para%c_hmc(i)%rho3            = input%hmc_rho3(i)
     para%c_hmc(i)%rho4            = input%hmc_rho4(i)
     para%c_hmc(i)%m_scale3        = input%hmc_m_scale3(i)
     para%c_hmc(i)%m_scale4        = input%hmc_m_scale4(i)
     para%c_hmc(i)%m_scale5        = input%hmc_m_scale5(i)
     para%c_hmc(i)%n_stout         = input%n_stout(i)
     para%c_hmc(i)%alpha           = input%alpha(i)
     para%c_hmc(i)%theta           = input%theta(i)

     read(para%c_hmc(i)%kappa_strange,   *) para%hmc(i)%kappa_strange
     read(para%c_hmc(i)%rho2,            *) para%hmc(i)%rho2
     read(para%c_hmc(i)%rho3,            *) para%hmc(i)%rho3
     read(para%c_hmc(i)%rho4,            *) para%hmc(i)%rho4
     read(para%c_hmc(i)%m_scale3,        *) para%hmc(i)%m_scale3
     read(para%c_hmc(i)%m_scale4,        *) para%hmc(i)%m_scale4
     read(para%c_hmc(i)%m_scale5,        *) para%hmc(i)%m_scale5

     read(para%c_hmc(i)%n_stout        ,*) para%hmc(i)%n_stout
     read(para%c_hmc(i)%alpha          ,*) para%hmc(i)%alpha
     read(para%c_hmc(i)%theta          ,*) para%hmc(i)%theta

     if (para%hmc(i)%kappa == ZERO .and. para%hmc(i)%csw /= ZERO) then
        para%hmc(i)%csw_kappa = para%hmc(i)%csw
        para%c_hmc(i)%csw = "-1 (infinity)"
        para%hmc(i)%csw = -1
        para%hmc(i)%kappa_strange = ZERO
        para%hmc(i)%csw_kappa_strange = ZERO
     else
        para%hmc(i)%csw_kappa         = para%hmc(i)%csw * para%hmc(i)%kappa
        para%hmc(i)%csw_kappa_strange = para%hmc(i)%csw * para%hmc(i)%kappa_strange
!        call check_csw(para%hmc(i)%beta,  para%hmc(i)%csw)          !%%%%%%%%%%%%%%nf2p1
     endif

     para%hmc(i)%tau = para%hmc(i)%traj_length / para%hmc(i)%ntau

     write(para%c_hmc(i)%csw_kappa, "(f20.15)") para%hmc(i)%csw_kappa
     write(para%c_hmc(i)%tau,       "(f20.15)") para%hmc(i)%tau
     write(para%c_hmc(i)%csw_kappa_strange, "(f20.15)") para%hmc(i)%csw_kappa_strange

     if (para%hmc(i)%kappa         == ZERO .and. &
         para%hmc(i)%kappa_strange == ZERO .and. &
         para%hmc(i)%csw           == ZERO) then
        quenched = .true.
     else
        dynamical = .true.
     endif

     if (para%hmc(i)%csw /= ZERO) clover     = .true.
     if (para%hmc(i)%h   /= ZERO) h_ext      = .true.

     para%hmc(i)%model = input%hmc_model
     para%hmc(i)%integrator(1) = input%hmc_integrator1
     para%hmc(i)%integrator(2) = input%hmc_integrator2
     para%hmc(i)%integrator(3) = input%hmc_integrator3
     para%hmc(i)%integrator(4) = input%hmc_integrator4
     para%hmc(i)%integrator(5) = input%hmc_integrator5
     para%hmc(i)%integrator(6) = input%hmc_integrator6

     !------------------------------------------- check model
     select case (para%hmc(i)%model)
        case ("A");call check_model_2p1( 1,para%hmc(i))
        case ("B");call check_model_2p1( 2,para%hmc(i))
        case ("C");call check_model_2p1( 3,para%hmc(i))
        case ("D");call check_model_2p1( 4,para%hmc(i))
!!        case ("E");call check_model_2p1( 5,para%hmc(i))
!!        case ("F");call check_model_2p1( 6,para%hmc(i))
!!        case default
!!           call die("init_para(): unknown model, model should be [A-F]")
     end select
     !------------------------------------------- check model

  enddo

  switches%mass_kappa = (input%hmc_hkappa /=0)
  switches%stout      = (para%hmc(1)%n_stout /= 0 .and. &
                             para%hmc(1)%alpha /= ZERO)
  switches%theta      = (para%hmc(1)%theta /= ZERO)
  switches%tadcsw     = (input%tadpole_csw /= 0)
  switches%fixtad     = (input%fix_coefficients_tadpole /= 0)
  switches%fixtad_u04 = (input%fix_coefficients_tadpole_by_u04 /=0)

  select case (input%start_configuration)
     case ("hot");  para%start = START_HOT
     case ("cold"); para%start = START_COLD
     case ("file"); para%start = START_FILE
     case default  
        call die("init_para(): start_configuration must be {hot|cold|file}")
  end select

  select case (input%start_random)
     case ("random");  para%seed = -1
     case ("default"); para%seed = 0
     case default;     read(input%start_random, *) para%seed
  end select

  select case (input%tempering_swap_sequence)
     case ("up");     para%swap_seq = SWAP_UP
     case ("down");   para%swap_seq = SWAP_DOWN
     case ("random"); para%swap_seq = SWAP_RANDOM
     case default
       call die("init_para(): tempering_swap_sequence must be {up|down|random}")
  end select

  switches%stout_ds = .false.
  switches%stout_cl = .false.
  select case (input%fermi_action)
     case ("NON")
        if (dynamical) call die("NON fermi_action but dynamical")
     case ("WILSON")
        if (clover) call die("WILSON fermi_action but csw /= 0")
     case ("CLOVER")
        if (.not. clover) call die("CLOVER fermi_action but csw = 0")
     case ("SLW")
        if (clover) call die("SLW fermi_action but csw /= 0")
        if (.not. switches%stout) &
        call die("SLW  fermi_action but not stout")
        switches%stout_ds = .true.
     case ("SLOC")
        if (.not. switches%stout) &
        call die("SLOC fermi_action but not stout")
        switches%stout_ds = .true.
        switches%stout_cl = .true.
     case ("SLIC")
        if (.not. switches%stout) &
        call die("SLIC fermi_action but not stout")
        switches%stout_cl = .true.
     case ("SLRC")
        if (.not. switches%stout) &
        call die("SLRC fermi_action but not stout")
        switches%stout_ds = .true.
     case default
        call die("init_para(): fermi_action must be {NON|WILSON|CLOVER|SLW|SLOC|SLIC|SLRC}")
  end select

  if (quenched .and. dynamical) call die("init_para(): quenched/dynamical mixed")

  if (para%nforce < 0) call die("init_para(): nforce < 0")

  if (flags%continuation_job) then
      para%start = START_CONT
      switches%start_continue = .true.
  endif

  switches%quenched   = quenched
  switches%dynamical  = dynamical
  switches%clover     = clover
  switches%h_ext      = h_ext
  switches%hasenbusch = (para%hmc(1)%rho /= 0)
  switches%rhmc   = (input%hmc_model == "F" )

  if (quenched) then
    switches%hasenbusch = .false.
  endif

  switches%tempering             = .false.
  switches%measure_polyakov_loop = .false.
  switches%measure_traces        = .false.

  if (input%ensembles             >  1) switches%tempering             = .true.
  if (input%measure_polyakov_loop /= 0) switches%measure_polyakov_loop = .true.
  if (input%measure_traces        /= 0) switches%measure_traces        = .true.

  switches%measurement_only = .false.
  switches%boundary_sf = .false.
  if (input%measurement_only   /= 0) switches%measurement_only = .true.

  if (input%boundary_sf   /= 0) switches%boundary_sf = .true.

  if (input%hmc_test == 0) then
     switches%hmc_test = .false.
  else
     switches%hmc_test = .true.
  endif


  call init_action(para)

end

!-------------------------------------------------------------------------------
subroutine init_counter(para, flags)

  use typedef_flags
  use typedef_para
  use module_counter
  use module_function_decl
  implicit none

  type(type_para)    :: para
  type(type_flags)   :: flags
  FILENAME, external :: count_file, stop_file

  if (f_exist(stop_file())) then
     call die("init_counter(): found stop file " // stop_file())
  endif

  counter%run = para%run
  counter%j_traj = 0

  if (flags%continuation_job) then
     open(UCOUNT, file = count_file(), action = "read", status = "old")
     read(UCOUNT, *) counter%run
     read(UCOUNT, *) counter%job
     read(UCOUNT, *) counter%traj
     close(UCOUNT)

     if (counter%run /= para%run) call die("init_counter(): RUN inconsistent")
     counter%job = counter%job + 1
  else
     counter%run = para%run
     counter%job = 1
     counter%traj = -para%nforce
  endif

end

!-------------------------------------------------------------------------------
subroutine write_counter(maxtraj)

  use module_counter
  use module_function_decl
  implicit none

  integer            :: maxtraj
  FILENAME, external :: count_file, stop_file

  if (my_pe() /= 0) return

  open(UCOUNT, file = count_file(), action = "write")
  write(UCOUNT, *) counter%run, " run"
  write(UCOUNT, *) counter%job, " job"
  write(UCOUNT, *) counter%traj, " traj"
  close(UCOUNT)

  if (counter%traj >= maxtraj) then
     open(UCOUNT, file = stop_file(), status = "unknown")
     close(UCOUNT)
  endif

end

!-------------------------------------------------------------------------------
subroutine write_header(para, flags)

  use typedef_flags
  use typedef_para
  use module_bqcd
  use module_counter
  use module_function_decl
  use module_input
  use module_mre
  use module_svn
  use module_thread
  use module_switches
  use module_action
  use module_dd
  implicit none

  type(type_para), intent(in)  :: para
  type(type_flags), intent(in) :: flags
  integer                      :: i, nts
  SEED                         :: seed
  character(len = 50)          :: fmt, tmp_c
  character(len = 4), external :: format_ensemble
  character(len = 50), external::c_info_mdsteps
  integer, external :: num_time_scale

  if (my_pe() == 0) then

     fmt = "(1x,a," // format_ensemble() // ",2a)"

     call begin(UREC, "Header")

    if (input%comment /= "") then
     write(UREC, 405) "Comment", trim(input%comment)
    endif

     write(UREC, 400) "Program", prog_name, prog_version // prog_revision
#ifdef BAGEL
     write(UREC,   *) "Version_of_D ", "BAGEL"
#else
     write(UREC,   *) "Version_of_D ", version_of_d()
#endif
     write(UREC,   *) "Communication ", trim(comm_method())
     write(UREC,   *) "RandomNumbers ", trim(ran_name())
     write(UREC,   *) "Run ", para%run
     write(UREC,   *) "Job ", counter%job
     write(UREC, 405) "Host", rechner()
     write(UREC, 400) "Date", datum(), uhrzeit()
     write(UREC, 410) "L               ", para%L
     write(UREC, 410) "DDL             ", input%ddlattice
     write(UREC, 410) "NPE             ", para%NPE
     write(UREC, 410) "process_mapping ", para%process_mapping
     write(UREC, 410) "bc_fermions     ", para%bc_fermions
     write(UREC, 410) "gamma_index     ", para%gamma_index
     write(UREC, 405) "Gauge Action         ", gauge_action
     write(UREC, 405) "Fermi Action         ", fermi_action

#ifdef OMTDTD
     write(UREC, 405) "tilde M      ", "1 - T^-1 D T^-1 D"
#else
     write(UREC, 405) "tilde M      ", "T - D T^-1 D"
#endif

#ifdef GAMMA_NOTATION_CHROMA
     write(UREC, 405) "Gamma notation       ", "Chroma"
#elif defined GAMMA_NOTATION_CHIRAL
     write(UREC, 405) "Gamma notation       ", "Chiral"
#elif defined GAMMA_NOTATION_DDHMC
     write(UREC, 405) "Gamma notation       ", "DDHMC"
#else
     write(UREC, 405) "Gamma notation       ",  "BQCD"
#endif

#ifdef DOEDEO
     write(UREC, *) "eo-prec DoeDeo (Chroma style)"
#else
     write(UREC, *) "eo-prec DeoDoe               "
#endif

#ifdef ZERO_MOMENTUM
     write(UREC, *) "momentum is zero"
#endif
#ifdef NON_REALITY
     write(UREC, *) "pfermion has non reality"
#endif

     if (switches%boundary_sf) then
        write(UREC, *) "BoundaryCondition   schr  "
     else
        write(UREC, *) "BoundaryCondition   normal"
     endif

     write(UREC,   *) "Threads ", n_thread
     write(UREC,   *) "Start   ", para%start

     if (para%start == START_FILE) then
        do i = 1, para%n_temper
           write(UREC, fmt) "StartConf_", i, " ", trim(para%info_file(i))
        enddo
     endif

     if (flags%continuation_job) then
        seed = ran_get_seed()
     else
        seed = para%seed
     endif

     write(UREC,   *) "Seed    ", seed
     write(UREC,   *) "Swap_seq", para%swap_seq
     write(UREC,   *) "N_force ", para%nforce
     write(UREC,   *) "N_traj  ", para%ntraj
     write(UREC,   *) "N_save  ", para%nsave
     write(UREC,   *) "N_temper", para%n_temper

     do i = 1, para%n_temper
        nts=num_time_scale(para%hmc(i))
        write(UREC, fmt) "beta_", i, "              ", trim(para%c_hmc(i)%beta)
        write(UREC, fmt) "c0_", i, "                ", trim(input%c0(i))
        write(UREC, fmt) "c1_", i, "                ", trim(input%c1(i))
        write(UREC, fmt) "c2_", i, "                ", trim(input%c2(i))
        write(UREC, fmt) "c3_", i, "                ", trim(input%c3(i))
        write(UREC, fmt) "u04_", i, "               ", trim(input%u04(i))
        write(UREC, fmt) "kappa_", i, "             ", trim(para%c_hmc(i)%kappa)
        write(UREC, fmt) "kappa_strange_", i, "     ", trim(para%c_hmc(i)%kappa_strange)
        write(UREC, fmt) "csw_", i, "               ", trim(para%c_hmc(i)%csw)
        write(UREC, fmt) "csw_kappa_", i, "         ", trim(para%c_hmc(i)%csw_kappa)
        write(UREC, fmt) "csw_kappa_strange_", i, " ", trim(para%c_hmc(i)%csw_kappa_strange)
        write(UREC, fmt) "n_stout_", i, "           ", trim(para%c_hmc(i)%n_stout)
        write(UREC, fmt) "alpha_", i, "             ", trim(para%c_hmc(i)%alpha)
        write(UREC, fmt) "theta_", i, "             ", trim(para%c_hmc(i)%theta)
        write(UREC, fmt) "h_", i, "                 ", trim(para%c_hmc(i)%h)
        write(UREC, fmt) "rho_", i, "               ", trim(para%c_hmc(i)%rho)
        write(UREC, fmt) "rho2_", i, "              ", trim(para%c_hmc(i)%rho2)
        write(UREC, fmt) "rho3_", i, "              ", trim(para%c_hmc(i)%rho3)
        write(UREC, fmt) "rho4_", i, "              ", trim(para%c_hmc(i)%rho4)
        write(UREC, fmt) "traj_length_", i, "       ", trim(para%c_hmc(i)%traj_length)
        write(UREC, fmt) "tau_", i, "               ", trim(para%c_hmc(i)%tau)
        write(UREC, fmt) "N_tau_", i, "             ", trim(c_info_mdsteps(para%c_hmc(i)%ntau,    1,para%hmc(i)))
if(nts>1)write(UREC, fmt) "m_scale_", i, "           ", trim(c_info_mdsteps(para%c_hmc(i)%m_scale ,2,para%hmc(i)))
if(nts>2)write(UREC, fmt) "m_scale2_", i, "          ", trim(c_info_mdsteps(para%c_hmc(i)%m_scale2,3,para%hmc(i)))
if(nts>3)write(UREC, fmt) "m_scale3_", i, "          ", trim(c_info_mdsteps(para%c_hmc(i)%m_scale3,4,para%hmc(i)))
if(nts>4)write(UREC, fmt) "m_scale4_", i, "          ", trim(c_info_mdsteps(para%c_hmc(i)%m_scale4,5,para%hmc(i)))
if(nts>5)write(UREC, fmt) "m_scale5_", i, "          ", trim(c_info_mdsteps(para%c_hmc(i)%m_scale5,6,para%hmc(i)))
     enddo
     write(UREC, *) "hkappa",          input%hmc_hkappa
     write(UREC, *) "mpf_eo[m,l,s]:",  input%hmc_mpf_mass, input%hmc_mpf_rhmc_l, input%hmc_mpf_rhmc_s
     write(UREC, *) "mpf_dd[m,l,s]:",  input%hmc_mpf_dd_mass, input%hmc_mpf_dd_rhmc_l, input%hmc_mpf_dd_rhmc_s
     write(UREC, *) "mpf_hh[m,l,s]:",  input%hmc_mpf_hh_mass, input%hmc_mpf_hh_rhmc_l, input%hmc_mpf_hh_rhmc_s
!!     write(UREC, *) "fraction_u1",     input%rhmc_fraction_u1
!!     write(UREC, *) "fraction_u2",     input%rhmc_fraction_u2
!!     write(UREC, *) "roughness_level", input%rhmc_roughness_level
     write(UREC,   *) "HMC_model ", para%hmc(1)%model
     write(UREC,   *) "REAL_kind ", RKIND
     write(UREC,   *) "Solver_outer ", trim(input%solver_outer_solver)
     write(UREC,   *) "Solver_inner ", trim(input%solver_inner_solver)
     write(UREC, 405) "CG_rest    ", trim(para%c_cg_rest)
     write(UREC, 405) "CG_rest_md ", trim(input%solver_rest_md)
     write(UREC,   *) "CG_stopping_criterion ", input%solver_stopping_criterion
     write(UREC,   *) "CG_outer_steps ", input%solver_outer_steps
     if (ddable) then
        write(UREC, 404) "DD_rest(mc md sap):   ",trim(input%solver_rest_domain_mc), &
                                                  trim(input%solver_rest_domain_md), &
                                                  trim(input%solver_rest_domain_sap)
        write(UREC, 410) "DD_num[in(kv df) out(kv sap)]:", &
                                               input%solver_numkrylov_mtdmtdd, &
                                               input%solver_numdefvec_mtdmtdd, &
                                               input%solver_defsapgcr_nkv, &
                                               input%solver_defsapgcr_nsp
        if (ddallactive) then
           write(UREC,   *) "DD: all active"
        else
           write(UREC,   *) "DD: halo is NOT active"
        endif
     endif

     write(UREC,   *) "MRE_vectors ", mre_n_vec
     if (input%fullsolver=="dd" .and. ddable) then
        write(UREC,   *) "Fullsolver DD"
     else
        write(UREC,   *) "Fullsolver eo"
     endif

     call end(UREC, "Header")

400  format (3(1x,a))
404  format (4(1x,a))
405  format (2(1x,a))
408  format (2(1x,a),1x,3a)
410  format (1x,a,4i3)

  endif

end

!-------------------------------------------------------------------------------
subroutine write_footer(time0)

  use module_function_decl
  use module_thread
  implicit none

  SEED    :: seed
  SECONDS :: time0, sekunden

  call begin(UREC, "Footer")

  if (my_pe() == 0) then
     write(UREC, 400) "Date", datum(), uhrzeit()
     write(UREC,   *) "Seed", ran_get_seed()
     write(UREC, 410) "CPU-Time", &
                      sekunden() - time0, "s on", num_pes() * n_thread, "CPUs"
  endif

400  format (3(1x,a))
410  format (1x,a,1x,f8.1,1x,a,1x,i5,1x,a)

  TIMING_WRITE(UREC)
  call end(UREC, "Footer")

end

!-------------------------------------------------------------------------------
subroutine get_flags(flags) ! read command line arguments [and open output file]

  use typedef_cksum
  use typedef_flags
  use module_bqcd
  use module_function_decl
  use module_input
  implicit none

  type(type_flags), intent(out) :: flags

  integer                       :: iarg, length, stat, narg
  integer, external             :: ipxfargc
  character(len = 2)            :: opt

  flags%continuation_job = .false.
  flags%show_version = .false.

  narg = ipxfargc()

  iarg = 1
  do while (iarg <= narg)
     call pxfgetarg(iarg, opt, length, stat)

     if (opt(1:1) == "-") then
        if (length > 2) call usage()

        select case (opt(2:2))
        case ("c")
           flags%continuation_job = .true.
           iarg = iarg + 1
        case ("I")
           call input_dump(6)
           call comm_finalize()
           stop
        case ("V")
           flags%show_version = .true.
           iarg = iarg + 1
        case default
           call usage
        end select
     else
        exit
     endif
  enddo
           
  if (flags%show_version) then
     call input_init()  ! some input defaults are needed in version()
     call version()
     call comm_finalize()
     stop
  endif

  call take_arg(iarg, flags%input, narg)

  if (iarg <= narg) then
     call take_arg(iarg, flags%output, narg)
     if (my_pe() == 0) then
        close(UREC)

        if (flags%continuation_job) then
           open(UREC, file = flags%output, &
                   action = "write", position = "append", status = "unknown")
        else
           open(UREC, file = flags%output,  action = "write", status ="unknown")
        endif
     endif
  endif
  if (narg >= iarg) call usage

CONTAINS

  subroutine usage()
    implicit none
    call die("Usage: " // prog_name // " [-c] [-I] [-V] input [output]")
  end subroutine usage

  subroutine version()
    implicit none

    if (my_pe() == 0) then
       write(6,*) "This is ", prog_name, " ", prog_version, prog_revision
       write(6,*) "   input format:    ", input_version
       write(6,*) "   conf info format:", conf_info_version
       write(6,*) "   MAX_TEMPER:      ", MAX_TEMPER
       write(6,*) "   REAL kind:       ", RKIND
       write(6,*) "   Version of D:    ", version_of_d()
       write(6,*) "   Communication:   ", trim(comm_method())
       write(6,*) "   RandomNumbers:   ", trim(ran_name())
    endif
  end subroutine version

  subroutine take_arg(iarg, arg, narg)
    implicit none
    integer, intent(inout)          :: iarg
    character(len = *), intent(out) :: arg
    integer, intent(in)             :: narg
    integer                         :: length, stat

    if (iarg > narg) call usage
    call pxfgetarg(iarg, arg, length, stat)
    if (length > len(arg)) then
       call die("get_flags(): " // arg // ": argument too long")
    endif
    iarg = iarg + 1
  end subroutine take_arg

end
!===============================================================================
!===============================================================================
subroutine check_model_2p1(model,para)
  use module_input
  use typedef_hmc
  implicit none

  type(hmc_para),   intent(in)    :: para
  integer::model

  logical,dimension(9,15)::care, agree !(models,check list)
  character(len =  1),save :: model_name(6)
  character(len = 16),save :: para_name(15)
  character(len = 45) :: msg
  integer::i,j
  REAL::input_value(15),expect_value(6,15)
!---------------------------------------------
data model_name( 1) /"A"/
data model_name( 2) /"B"/
data model_name( 3) /"C"/
data model_name( 4) /"D"/
data model_name( 5) /"E"/
data model_name( 6) /"F"/

data para_name( 1) /"mpf_mass"/
data para_name( 2) /"mpf_rhmc_l"/
data para_name( 3) /"mpf_rhmc_s"/
data para_name( 4) /"rho"/
data para_name( 5) /"rho2"/
data para_name( 6) /"rho3"/
data para_name( 7) /"rho4"/
data para_name( 8) /"m_scale2"/
data para_name( 9) /"m_scale3"/
data para_name(10) /"m_scale4"/
data para_name(11) /"m_scale5"/
data para_name(12) /"kappa_strange"/
data para_name(13) /"n_stout"/
data para_name(14) /"alpha"/
data para_name(15) /"theta"/

input_value( 1)=input%hmc_mpf_mass
input_value( 2)=input%hmc_mpf_rhmc_l
input_value( 3)=input%hmc_mpf_rhmc_s
input_value( 4)=para%rho
input_value( 5)=para%rho2
input_value( 6)=para%rho3
input_value( 7)=para%rho4
input_value( 8)=para%m_scale2
input_value( 9)=para%m_scale3
input_value(10)=para%m_scale4
input_value(11)=para%m_scale5
input_value(12)=para%kappa_strange
input_value(13)=para%n_stout
input_value(14)=para%alpha
input_value(15)=para%theta
!------------------------------------------

care=  .true.
agree= .true.

do i=1,6
   expect_value(i, 1)=ONE
   expect_value(i, 2)=ZERO
   expect_value(i, 3)=ZERO
   expect_value(i, 4)=ZERO
   expect_value(i, 5)=ZERO
   expect_value(i, 6)=ZERO
   expect_value(i, 7)=ZERO
   expect_value(i, 8)=ONE
   expect_value(i, 9)=ONE
   expect_value(i,10)=ONE
   expect_value(i,11)=ONE
   expect_value(i,12)=ZERO
   expect_value(i,13)=ZERO
   expect_value(i,14)=ZERO
   expect_value(i,15)=ZERO
enddo

do i=2,5
   expect_value(i, 1)=TWO
enddo

if(para%kappa == ZERO .and. para%kappa_strange == ZERO)then
do i=1,6
   expect_value(i, 1)=ZERO
   expect_value(i, 2)=ZERO
   expect_value(i, 3)=ZERO
enddo
endif

care( 5:6, 1)= .false.
care( 5:6, 2)= .false.
care( 5:6, 3)= .false.

agree( 2, 4)= .false.
agree( 3, 4)= .false.
agree( 4, 4)= .false.
agree( 5, 4)= .false.
agree( 6, 4)= .false.

agree( 4, 8)= .false.
agree( 5, 8)= .false.
agree( 6, 8)= .false.

care( 5:6, 9)= .false.
care( 5:6,10)= .false.
care( 5:6,11)= .false.

agree( 6,12)= .false.

care(6, 13)= .false.
care(6, 14)= .false.

care(5, 15)= .false.
care(6, 15)= .false.

do i=1,15
   if ( care(model,i) ) then
      if ( agree(model,i) ) then
         if ( input_value(i) /= expect_value(model,i) ) then
            msg = " model "//model_name(model)//" was selected, but " &
                  //trim(para_name(i))//" /= "//char(48+int(expect_value(model,i)))
            call warn(msg)
         endif
      else
         if ( input_value(i) == expect_value(model,i) ) then
            msg = " model "//model_name(model)//" was selected, but " &
                  //trim(para_name(i))//" = "//char(48+int(expect_value(model,i)))
            call warn(msg)
         endif
      endif
   endif
enddo

end
!-------------------------------------------------------------------------------
character (len=50) function c_info_mdsteps(c_nsteps,nscale,parah)
  use module_input
  use typedef_hmc
  implicit none
  type(hmc_para), intent(in)    :: parah
  integer, intent(in) :: nscale
  character(len=*), intent(in) :: c_nsteps
  character(len=50) :: tmp_c, tmp_c1

  write(tmp_c,"(a3,a7)")trim(c_nsteps), parah%integrator(nscale)       ;tmp_c1=tmp_c
if(input%hmc_dsf1_mtmp==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "eo1";tmp_c1=tmp_c
if(input%hmc_dsf2_mtmp==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "eo2";tmp_c1=tmp_c
if(input%hmc_dsf3_mtmp==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "eo3";tmp_c1=tmp_c
if(input%hmc_dsf4_mtmp==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "eo4";tmp_c1=tmp_c
if(input%hmc_dsf5_mtmp==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "eo5";tmp_c1=tmp_c
if(input%hmc_dsf1_un  ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "un1";tmp_c1=tmp_c
if(input%hmc_dsf2_un  ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "un2";tmp_c1=tmp_c
if(input%hmc_dsf1_dd  ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "dd1";tmp_c1=tmp_c
if(input%hmc_dsf2_dd  ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "dd2";tmp_c1=tmp_c
if(input%hmc_dsf1_hh  ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "hh1";tmp_c1=tmp_c
if(input%hmc_dsf2_hh  ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "hh2";tmp_c1=tmp_c
if(input%hmc_dsf_erl  ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "erl";tmp_c1=tmp_c
if(input%hmc_dsf_drl  ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "drl";tmp_c1=tmp_c
if(input%hmc_dsf_hrl  ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "hrl";tmp_c1=tmp_c
if(input%hmc_dsf_ers  ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "ers";tmp_c1=tmp_c
if(input%hmc_dsf_drs  ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "drs";tmp_c1=tmp_c
if(input%hmc_dsf_hrs  ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "hrs";tmp_c1=tmp_c
!!if(input%hmc_dsfr     ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c), " fr"
if(input%hmc_dsd      ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "dat";tmp_c1=tmp_c
if(input%hmc_dsg      ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), "  g";tmp_c1=tmp_c
if(input%hmc_dsig     ==nscale)write(tmp_c,"(a,a4)")trim(tmp_c1), " ig";tmp_c1=tmp_c

  c_info_mdsteps=tmp_c1
end
!===============================================================================
