!===============================================================================
!
! conf.F90 - operations on gauge field and pseudo fermion field configurations
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 1998-2007 Hinnerk Stueben
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine init_confs(para, conf)

  use typedef_para
  use module_p_interface
  use module_switches
  implicit none

  type(type_para)                       :: para
  type(hmc_conf), dimension(MAX_TEMPER) :: conf
  integer                               :: i

  do i = 1, para%n_temper
     call allocate_g_field(conf(i)%u)
     call allocate_sc_field(conf(i)%phi)
     if (switches%hasenbusch) call allocate_sc_field(conf(i)%phi2)
     if (para%start == START_HOT .or. para%start == START_COLD) then
        call init_u(conf(i)%u, para%start)
        conf(i)%former = i
     endif
  enddo

  if (para%start == START_CONT) call conf_read(.true., para, conf)
  if (para%start == START_FILE) call conf_read(.false., para, conf)

  do i = 1, para%n_temper
     if (para%hmc(i)%csw_kappa /= ZERO) then
        call allocate_clover_field_a(conf(i)%a)
        call allocate_clover_field_a(conf(i)%i)
        call allocate_clover_field_b(conf(i)%b)
        call clover_init(conf(i)%a, conf(i)%i, conf(i)%b, &
                         conf(i)%u, para%hmc(i)%csw_kappa)
     endif
  enddo

end

!-------------------------------------------------------------------------------
subroutine init_u(u, start)  ! initialization of u-field (at trajectory 0)

  use module_vol
  implicit none

  GAUGE_FIELD :: u
  integer :: start

  select case (start)
     case (START_HOT)
        call conf_hot(u)
     case (START_COLD)
        call conf_cold(u)
     case default
        call die("init_u(): don't know how to start")
  end select

  call xbound_g_field(u)
  call d_g_init(u)
end

!-------------------------------------------------------------------------------
subroutine conf_check(u)  ! checks if u-field is SU(3)

  use module_vol
  implicit none

  GAUGE_FIELD, intent(in) :: u
  SU3 :: v
  SU3, parameter :: su3_one = reshape( &
                           (/ ONE,ZERO,ZERO, &
                              ZERO,ONE,ZERO, &
                              ZERO,ZERO,ONE /), &
                           (/ NCOL, NCOL /))
  REAL :: dev, d
  integer :: i, j, k, eo, mu


  dev = ZERO
  do mu = 1, DIM
     do eo = EVEN, ODD
        do k = 1, VOLH
           call su3_check_det(u(1, 1, k, eo, mu))
           call uud(v, u(1, 1, k, eo, mu), u(1, 1, k, eo, mu))
           do i = 1, NCOL
              do j = 1, NCOL
                 d = abs(Re(v(i, j)) - Re(su3_one(i, j))) & 
                   + abs(Im(v(i, j)) - Im(su3_one(i, j)))
              enddo
           enddo
           dev = max(dev, d)
        enddo
     enddo
  enddo

  if (dev > 1e-13) call die('conf_check(): dev > 1e-13')
!!write(0,'(x,a,e10.2)') 'conf_check(): max deviation is ', dev

end

!-------------------------------------------------------------------------------
subroutine conf_normalize(u)  ! normalizes u-field to SU(3)

  use module_vol
  implicit none

  GAUGE_FIELD, intent(inout) :: u
  integer :: i, eo, mu

  do mu = 1, DIM
     do eo = EVEN, ODD
        do i = 1, volh
           call u_normalize(u(1, 1, i, eo, mu))
        enddo
     enddo
  enddo

end

!-------------------------------------------------------------------------------
subroutine conf_zero(u)  ! init ("OpenMP first touch")

  use module_vol
  implicit none

  GAUGE_FIELD :: u
  integer :: i, eo, mu

  do mu = 1, DIM
     do eo = EVEN, ODD
        !$omp parallel do
        do i = 1, volh
           u(1, 1, i, eo, mu) = ZERO
           u(2, 1, i, eo, mu) = ZERO
           u(3, 1, i, eo, mu) = ZERO
           u(1, 2, i, eo, mu) = ZERO
           u(2, 2, i, eo, mu) = ZERO
           u(3, 2, i, eo, mu) = ZERO
           u(1, 3, i, eo, mu) = ZERO
           u(2, 3, i, eo, mu) = ZERO
           u(3, 3, i, eo, mu) = ZERO
        enddo
     enddo
  enddo

end

!-------------------------------------------------------------------------------
subroutine conf_cold(u)  ! cold start

  use module_vol
  implicit none

  GAUGE_FIELD :: u
  integer :: i, eo, mu

  do mu = 1, DIM
     do eo = EVEN, ODD
        do i = 1, volh
           u(1, 1, i, eo, mu) = ONE
           u(2, 1, i, eo, mu) = ZERO
           u(3, 1, i, eo, mu) = ZERO
           u(1, 2, i, eo, mu) = ZERO
           u(2, 2, i, eo, mu) = ONE
           u(3, 2, i, eo, mu) = ZERO
           u(1, 3, i, eo, mu) = ZERO
           u(2, 3, i, eo, mu) = ZERO
           u(3, 3, i, eo, mu) = ONE
        enddo
     enddo
  enddo

end

!-------------------------------------------------------------------------------
subroutine conf_hot(u)  ! hot start

  use module_vol
  implicit none

  GAUGE_FIELD :: u
  integer :: i, eo, mu

  do mu = 1, DIM
     do eo = EVEN, ODD
        call ran_gauss_volh(NCOL * NCOL, u(1, 1, 1, eo, mu), ONE, eo)
        do i = 1, volh
           call u_normalize(u(1, 1, i, eo, mu))
        enddo
     enddo
  enddo

end

!-------------------------------------------------------------------------------
subroutine conf_read(restart, para, conf)

  use typedef_para
  use module_input
  implicit none

  logical                               :: restart
  type(type_para)                       :: para
  type(hmc_conf), dimension(MAX_TEMPER) :: conf


  if (restart) then  ! continue a job chain: 
                     ! read restart file(s) written by BQCD

     if (input%io_restart_format == "bqcd") then

        call conf_read_bqcd(restart, para, conf)

     elseif (input%io_restart_format == "ildg") then

        call conf_read_ildg(restart, para, conf)

     else
        call die("conf_read(): 'io_restart_format' unknown")  
     endif

  else  ! start a job chain: 
        ! read existing file(s) possibly written by other programs

     if (input%start_ildg_file /= "") then

        call conf_read_ildg(restart, para, conf)

     elseif (input%start_info_file(1) /= "") then

        call conf_read_bqcd(restart, para, conf)

     else
        call die("conf_read(): no start file given")  
     endif

  endif
end

!-------------------------------------------------------------------------------
subroutine conf_write(restart, para, conf)

  use typedef_para
  use module_input
  implicit none

  logical                               :: restart
  type(type_para)                       :: para
  type(hmc_conf), dimension(MAX_TEMPER) :: conf


  if (restart) then
     if (input%io_restart_format == "bqcd") then

        call conf_write_bqcd(restart, para, conf)

     elseif (input%io_restart_format == "ildg") then

        call conf_write_ildg(restart, para, conf)

     else
        call die("conf_write(): 'io_restart_format' unknown")  
     endif

  else

     if (input%io_conf_format == "bqcd") then

        call conf_write_bqcd(restart, para, conf)

     elseif (input%io_conf_format == "ildg") then

        call conf_write_ildg(restart, para, conf)

     else
        call die("conf_write(): 'io_conf_format' unknown")  
     endif

  endif
end

!===============================================================================
