!===============================================================================
! 
! fermi_action.F90  -  initialises phi and calculates actions
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2008-2010 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
module module_sf
  REAL, dimension(:), pointer, save  :: sf
end
!-------------------------------------------------------------------------------
subroutine calc_fermi_action(init)
  use module_action
  use module_field
  use module_sf
  implicit none
  logical, intent(in) :: init
  integer :: i
  integer, save :: count = 0

  if (.not. associated(sf)) return
  DEBUG2S("Start: fermi_action")
  if (count == 0 .or. (.not. init)) call gene_new_coeff()
  count = 1

  do i = 1, size(sf)
     select case(action%fermi(i)%type)
     case(EO_NORMAL)  ; call fermi_action_massprec(         i,mpf(i)%sc, sf(i),init)
     case(DD_NORMAL)  ; call fermi_action_mtdagmt_dd(       i,mpf(i)%sc, sf(i),init)
     case(HH_NORMAL)  ; call fermi_action_mtdagmt_hh(       i,mpf(i)%sc, sf(i),init)
     case(EO_RATIONAL); call fermi_action_rhmc(EO_RATIONAL, i,mpf(i)%sc, sf(i),init)
     case(DD_RATIONAL); call fermi_action_rhmc(DD_RATIONAL, i,mpf(i)%sc, sf(i),init)
     case(HH_RATIONAL); call fermi_action_rhmc(HH_RATIONAL, i,mpf(i)%sc, sf(i),init)
     !! MAY BE REMOVED
     case(UN_NORMAL)  ; call fermi_action_unprec_massprec(    i,mpf(i)%sc_e,mpf(i)%sc_o,sf(i),init)
     case(UN_RATIONAL); call die(": UN_RATIONAL does not work")
     !! FOR TEST, MAY BE REMOVED
     case(HH_NORMAL_DD); call fermi_action_mtdagmt_hhdd(      i,mpf(i)%ddsc,sf(i),init)
     case(HH_NORMAL_BB); call fermi_action_mtdagmt_hhbb(      i,mpf(i)%ddsc,sf(i),init)
     end select
  enddo
  DEBUG2S("End: fermi_action")
end

!-------------------------------------------------------------------------------
subroutine calc_fermi_action_rhmc()
  use module_action
  use module_field
  use module_sf
  implicit none
  integer :: i

  if (.not. associated(sf)) return
  call gene_new_coeff()
  do i = 1, size(sf)
     select case(action%fermi(i)%type)
     case(EO_RATIONAL); call fermi_action_rhmc(EO_RATIONAL, i,mpf(i)%sc, sf(i),.false.)
     case(DD_RATIONAL); call fermi_action_rhmc(DD_RATIONAL, i,mpf(i)%sc, sf(i),.false.)
     case(HH_RATIONAL); call fermi_action_rhmc(HH_RATIONAL, i,mpf(i)%sc, sf(i),.false.)
     case(UN_RATIONAL); call die(": UN_RATIONAL does not work")
     end select
  enddo
end

!-------------------------------------------------------------------------------
subroutine fermi_action_massprec(fid, inout, sf, init)
  use module_action
  use module_function_decl
  use module_p_interface
  use module_vol
  implicit none
  COMPLEX, intent(inout) :: inout(NDIRAC, NCOL, volh)

!!!!!!  SPINCOL_FIELD, intent(inout) :: inout
  REAL,          intent(out) :: sf
  logical,       intent(in)  :: init
  integer,       intent(in)  :: fid
  integer                    :: mid1, mid2, ierr
  COMPLEX, allocatable :: tmp1(:,:,:), tmp2(:,:,:)

  DEBUG2S("Start: fermi_action_massprec")
  mid1 = action%fermi(fid)%mid1
  mid2 = action%fermi(fid)%mid2

  allocate(tmp1(NDIRAC, NCOL, volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("fermi_action_massprec: allocation failed tmp1")
  endif
  allocate(tmp2(NDIRAC, NCOL, volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("fermi_action_massprec: allocation failed tmp2")
  endif

  if (init) then
     call random_sc_field(tmp1)
     sf = dotprod(tmp1, tmp1, SIZE_SC_FIELD)
     call init_quark(mid1)
     call wdag_r8(tmp2, tmp1, mid1)
     if (mid2 /= 0) then
        call init_quark(mid2)
        call cg_outer(tmp1, tmp2, mid2, 0, CG_MC)
        call wmul_r8(tmp2, tmp1, mid2)
     endif
     call sc_copy(inout,tmp2)
  else
     if (mid2 /= 0) then
        call init_quark(mid2)
        call sc_copy(tmp1, inout)
        call wdag_r8(tmp2, tmp1, mid2)
     else
        call sc_copy(tmp2, inout)
     endif
     !! tmp1=tmp2 is needed, since cg_outer has a preconditioning
     tmp1=tmp2
     call init_quark(mid1)
     call cg_outer(tmp1, tmp2, mid1, 0, CG_MC)
     call wmul_r8(tmp2, tmp1, mid1)
     sf = dotprod(tmp2, tmp2, SIZE_SC_FIELD)
  endif

  deallocate(tmp1, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("fermi_action_massprec: deallocation failed tmp1")
  endif
  deallocate(tmp2, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("fermi_action_massprec: deallocation failed tmp2")
  endif

  DEBUG2S("End: fermi_action_massprec")
end

!-------------------------------------------------------------------------------
subroutine fermi_action_mtdagmt_dd(fid, inout, sf, init)
  use module_action
  use module_function_decl
  use module_dd
  use module_vol
  implicit none
  COMPLEX, dimension(NDIRAC,NCOL,ld%lat%volh,0:ld%lat%num_ip-1),intent(inout) :: inout
  COMPLEX, dimension(NDIRAC,NCOL,ld%lat%volh,0:ld%lat%num_ip-1) :: tmp,tmp1
  REAL,          intent(out) :: sf
  logical,       intent(in)  :: init
  integer,       intent(in)  :: fid
  integer                    :: mid1, mid2

  DEBUG2S("Start: fermi_action_mtdagmt_dd")
  mid1 = action%fermi(fid)%mid1
  mid2 = action%fermi(fid)%mid2

  if (init) then
     call random_sc_field_dd(tmp)
     sf = dotprod(tmp, tmp, SIZE_SC_FIELD)
     call init_quark(mid1)
     call wtdag_all_dd(inout, tmp, mid1)
     if (mid2 /= 0) then
        call init_quark(mid2)
        call cg_all_dd_outer(tmp,inout,mid2,0,CG_MC)
        call wtmul_all_dd(inout, tmp, mid2)
     endif
  else
     if (mid2 /= 0) then
        call init_quark(mid2)
        call wtdag_all_dd(tmp, inout, mid2)
     else
        tmp=inout
     endif
     call init_quark(mid1)
     tmp1=tmp
     call cg_all_dd_outer(tmp1,tmp,mid1,0,CG_MC)
     call wtmul_all_dd(tmp,tmp1,mid1)
     sf = dotprod(tmp, tmp, SIZE_SC_FIELD)
  endif
  DEBUG2S("End: fermi_action_mtdagmt_dd")
end
!-------------------------------------------------------------------------------
subroutine fermi_action_mtdagmt_hh(fid, inout, sf, init)
  use module_action
  use module_function_decl
  use module_dd
  use module_vol
  implicit none
  COMPLEX, dimension(NDIRAC,NCOL,ld%lat%volh,0:ld%lat%num_ip-1),intent(inout) :: inout
  COMPLEX, dimension(NDIRAC,NCOL,ld%lat%volh,0:ld%lat%num_ip-1) :: tmp,tmp1
  REAL,          intent(out) :: sf
  logical,       intent(in)  :: init
  integer,       intent(in)  :: fid
  integer                    :: mid1, mid2

  DEBUG2S("Start: fermi_action_mtdagmt_hh")
  mid1 = action%fermi(fid)%mid1
  mid2 = action%fermi(fid)%mid2

  if (init) then
     call random_sc_field_hh(tmp)
     sf = dotprod(tmp, tmp, SIZE_SC_FIELD)
     call init_quark(mid1)
     call unprec_hdag(inout,  tmp,mid1)
     if (mid2 /= 0) then
        call init_quark(mid2)
        call cg_hh_outer(tmp,  inout,mid2, 0, CG_MC)
        call unprec_hmul(inout,  tmp,mid2)
     endif
  else
     if (mid2 /= 0) then
        call init_quark(mid2)
        call unprec_hdag(tmp,inout, mid2)
     else
        tmp=inout
     endif
     call init_quark(mid1)
     tmp1=tmp
     call cg_hh_outer( tmp1,  tmp,mid1, 0, CG_MC)
     call unprec_hmul(  tmp, tmp1,mid1)
     sf = dotprod(tmp, tmp, SIZE_SC_FIELD)
  endif
  DEBUG2S("End: fermi_action_mtdagmt_hh")
end

!-------------------------------------------------------------------------------
subroutine fermi_action_rhmc(type, fid, inout, sf, init)
  use module_action
  use module_function_decl
  use module_vol
  use module_rhmc
  implicit none

  COMPLEX, intent(inout) :: inout(NDIRAC, NCOL, volh)
!!  SPINCOL_FIELD, intent(inout) :: inout
  REAL,          intent(out)   :: sf
  logical,       intent(in)    :: init
  integer,       intent(in)    :: fid, type
  integer                      :: mid, rid, ierr
  COMPLEX, allocatable :: tmp(:,:,:)
  REAL :: factor, exponent

  DEBUG2S("Start : fermi_action_rhmc")
  mid = action%fermi(fid)%mid1

  allocate(tmp(NDIRAC, NCOL, volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("fermi_action_rhmc: allocation failed tmp")
  endif

  call init_quark(mid)

  if (init) then
!     factor= ONE
!     exponent= ZERO
     factor= action%fermi(fid)%factor
     rid   = action%fermi(fid)%rid
     exponent= - ratapp(rid)%exponents(2)
     select case(type)
       case(EO_RATIONAL); call random_sc_field(tmp)
       case(DD_RATIONAL); call random_sc_field_dd(tmp)
       case(HH_RATIONAL); call random_sc_field_hh(tmp)
     end select
     sf =dotprod(tmp, tmp, SIZE_SC_FIELD)
     call rhmc_action(type, tmp, mid, ratapp(rid)%shift_exact, ratapp(rid)%abar_exact_p, ratapp(rid)%bbar_exact_p, factor, exponent)
     call sc_copy(inout, tmp)
  else
!     factor= ONE
!     exponent= ZERO
     factor= action%fermi(fid)%factor
     rid   = action%fermi(fid)%rid
     exponent= ratapp(rid)%exponents(2)
     call sc_copy(tmp, inout)
     call rhmc_action(type, tmp, mid, ratapp(rid)%shift_exact, ratapp(rid)%abar_exact_m, ratapp(rid)%bbar_exact_m, factor, exponent)
     sf =dotprod(tmp, tmp, SIZE_SC_FIELD)
  endif

  deallocate(tmp, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("fermi_action_rhmc: deallocation failed tmp")
  endif

  DEBUG2S("End: fermi_action_rhmc")

end

!-------------------------------------------------------------------------------
subroutine rhmc_action(type, inout, mid, n, a, b, factor, exponent)
  use module_vol
  use module_field
  implicit none

  SPINCOL_FIELD, intent(inout) :: inout
  integer,       intent(in)    :: mid, n, type
  REAL,          intent(in)    :: a(0:n), b(n), factor, exponent
  integer :: frdo(n), iters(n), i
  REAL :: fe(0:n), bof(n)

  fe(0)=factor**exponent * a(0)
  fe(1:n)=factor**exponent * a(1:n) / factor
  bof=b/factor
  frdo =0

  call mcg_outer(type, inout, mid, n, frdo, bof, iters, 0, CG_MC)
  call sc_scale(inout, fe(0))
  do i=1, n
     call sc_axpy(inout, sco(1,i)%sco(1), fe(i))
  enddo

end

!-------------------------------------------------------------------------------
subroutine random_sc_field(out)
  use module_vol
  implicit none
  SPINCOL_FIELD, intent(out) :: out

#ifdef NON_REALITY
  out=cmplx(dsqrt(0.3_8),dsqrt(0.2_8),kind=8)
#else
  call ran_gauss_volh(NDIRAC * NCOL, out, HALF, EVEN)
#endif

#ifdef FLIPSC
  call scfield_reorder_43_to_34(out)
#endif

end

!-------------------------------------------------------------------------------
subroutine random_sc_field_dd(out)
  use module_vol
  implicit none
  SPINCOL_FIELD, intent(out) :: out
  SPINCOL_FIELD :: tmp

#ifdef NON_REALITY
  tmp=cmplx(dsqrt(0.3_8),dsqrt(0.2_8),kind=8)
#else
  call ran_gauss_volh(NDIRAC * NCOL, tmp, HALF, EVEN)
#endif
  call sc_realorder(out,tmp)
end

!-------------------------------------------------------------------------------
subroutine random_sc_field_hh(out)
  use module_dd
  implicit none
  COMPLEX, dimension(4,3,ld%lat%volh,0:ld%lat%num_ip-1),intent(out) :: out
  COMPLEX, dimension(4,3,ld%lat%volh,0:ld%lat%num_ip-1) :: tmp
  COMPLEX, dimension(4,3,ld%lat%volh,0:ld%lat%num_ip-1,0:1) :: ddsc

#ifdef NON_REALITY
  out=cmplx(dsqrt(0.3_8),dsqrt(0.2_8),kind=8)
  tmp=cmplx(dsqrt(0.3_8),dsqrt(0.2_8),kind=8)
#else
  call ran_gauss_volh(NDIRAC * NCOL, out, HALF, EVEN)
  call ran_gauss_volh(NDIRAC * NCOL, tmp, HALF, ODD)
#endif

  call sc2ddsc(ddsc,out,tmp)
  call ddsc_dcmp(out(1,1,1,0              ),ddsc(1,1,1,0,0),0)
  call ddsc_dcmp(out(1,1,1,ld%lat%num_ip/2),ddsc(1,1,1,0,1),0)

end

!===============================================================================
!
! below could be removed
!
!-------------------------------------------------------------------------------
subroutine fermi_action_unprec_massprec(fid, inoute, inouto, sf, init)
  use module_action
  use module_function_decl
  use module_p_interface
  use module_vol
  implicit none
  SPINCOL_FIELD, intent(inout) :: inoute, inouto
  REAL,          intent(out) :: sf
  logical,       intent(in)  :: init
  integer,       intent(in)  :: fid
  integer                    :: mid1, mid2, ierr
  COMPLEX, allocatable :: tmpe(:,:,:)
  COMPLEX, allocatable :: tmp1e(:,:,:)
  COMPLEX, allocatable :: tmpo(:,:,:)
  COMPLEX, allocatable :: tmp1o(:,:,:)

  DEBUG2S("Start: fermi_action_unprec_massprec")
  mid1 = action%fermi(fid)%mid1
  mid2 = action%fermi(fid)%mid2

  allocate(tmpe(NDIRAC, NCOL, volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("fermi_action_massprec: allocation failed tmpe")
  endif
  allocate(tmp1e(NDIRAC, NCOL, volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("fermi_action_massprec: allocation failed tmp1e")
  endif
  allocate(tmpo(NDIRAC, NCOL, volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("fermi_action_massprec: allocation failed tmpo")
  endif
  allocate(tmp1o(NDIRAC, NCOL, volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("fermi_action_massprec: allocation failed tmp1o")
  endif


  if (init) then
     call random_sc_field(tmpe)
     call random_sc_field(tmpo)
     sf = dotprod(tmpe, tmpe, SIZE_SC_FIELD) + dotprod(tmpo, tmpo, SIZE_SC_FIELD)
     call init_quark(mid1)
     call unprec_mdag(inoute, inouto, tmpe, tmpo, mid1)
     if (mid2 /= 0) then
        call init_quark(mid2)
        call cg_unprec_outer(tmpe, tmpo, inoute, inouto, mid2, 0, CG_MC)
        call unprec_mmul(inoute, inouto, tmpe, tmpo, mid2)
     endif
  else
     if (mid2 /= 0) then
        call init_quark(mid2)
        call unprec_mdag(tmpe, tmpo, inoute, inouto, mid2)
     else
        call sc_copy(tmpe, inoute)
        call sc_copy(tmpo, inouto)
     endif
     call init_quark(mid1)
     tmp1e=tmpe
     tmp1o=tmpo
     call cg_unprec_outer(tmp1e, tmp1o, tmpe, tmpo, mid1, 0, CG_MC)
     call unprec_mmul(tmpe, tmpo, tmp1e, tmp1o, mid1)
     sf = dotprod(tmpe, tmpe, SIZE_SC_FIELD) + dotprod(tmpo, tmpo, SIZE_SC_FIELD)
  endif

  deallocate(tmpe, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("fermi_action_massprec: deallocation failed tmpe")
  endif
  deallocate(tmp1e, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("fermi_action_massprec: deallocation failed tmp1e")
  endif
  deallocate(tmpo, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("fermi_action_massprec: deallocation failed tmpo")
  endif
  deallocate(tmp1o, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("fermi_action_massprec: deallocation failed tmp1o")
  endif

  DEBUG2S("End: fermi_action_unprec_massprec")
end

!-------------------------------------------------------------------------------
subroutine fermi_action_mtdagmt_hhdd(fid, inout, sf, init)
  use module_action
  use module_function_decl
  use module_dd
  use module_vol
  implicit none
  COMPLEX, dimension(NDIRAC,NCOL,ld%lat%volh,0:ld%lat%num_ip-1,0:1),intent(inout) :: inout
  COMPLEX, dimension(NDIRAC,NCOL,ld%lat%volh,0:ld%lat%num_ip-1,0:1) :: tmp, tmp1
  REAL,          intent(out) :: sf
  logical,       intent(in)  :: init
  integer,       intent(in)  :: fid
  integer                    :: mid1, mid2, ip

  DEBUG2S("Start: fermi_action_mtdagmt_hhdd")
  mid1 = action%fermi(fid)%mid1
  mid2 = action%fermi(fid)%mid2


  if (init) then
     call random_sc_field(tmp(1,1,1,0,0))
     call random_sc_field(tmp(1,1,1,0,1))
     do ip = 0, ld%lat%num_ip-1
        if (eo_dm(ip) == 1) then
           tmp(1:4,1:3,1:ld%lat%volh,ip,0:1)=0
        endif
     enddo
     sf = dotprod(tmp, tmp, 2* SIZE_SC_FIELD)
     call init_quark(mid1)
     call unprec_mdag_dd(inout,  tmp,mid1)

  else
     tmp1=inout
     call init_quark(mid1)
     call cg_unprec_mdagm_dd_outer( tmp1,inout,mid1,          0, CG_MC)
     call unprec_mmul_dd(  tmp, tmp1,mid1)
     do ip = 0,ld%lat%num_ip-1
        if (eo_dm(ip) == 1) then
           tmp(1:4,1:3,1:ld%lat%volh,ip,0:1)=0
        endif
     enddo
     sf = dotprod(tmp, tmp, 2* SIZE_SC_FIELD)
  endif
  DEBUG2S("End: fermi_action_mtdagmt_hhdd")
end
!-------------------------------------------------------------------------------
subroutine fermi_action_mtdagmt_hhbb(fid, inout, sf, init)
  use module_action
  use module_function_decl
  use module_dd
  use module_vol
  implicit none
  COMPLEX, dimension(NDIRAC,NCOL,ld%lat%volh,0:ld%lat%num_ip-1,0:1),intent(inout) :: inout
  COMPLEX, dimension(NDIRAC,NCOL,ld%lat%volh,0:ld%lat%num_ip-1,0:1) :: tmp, tmp1
  REAL,          intent(out) :: sf
  logical,       intent(in)  :: init
  integer,       intent(in)  :: fid
  integer                    :: mid1, mid2, ip

  DEBUG2S("Start: fermi_action_mtdagmt_hhbb")
  mid1 = action%fermi(fid)%mid1
  mid2 = action%fermi(fid)%mid2

  if (init) then
     call random_sc_field(tmp(1,1,1,0,0))
     call random_sc_field(tmp(1,1,1,0,1))
     do ip = 0, ld%lat%num_ip-1
        if (eo_dm(ip) == 0) then
           tmp(1:4,1:3,1:ld%lat%volh,ip,0:1)=0
        endif
     enddo
     sf = dotprod(tmp, tmp, 2* SIZE_SC_FIELD)
     call init_quark(mid1)
     call unprec_mdag_bb(inout,  tmp,mid1,0 )
  else
     tmp1=inout
     call init_quark(mid1)
     call cg_unprec_mdagm_bb_outer( tmp1,inout,mid1,          0, CG_MC)
     call unprec_mmul_bb(  tmp, tmp1,mid1, 1)
     do ip = 0,ld%lat%num_ip-1
        if (eo_dm(ip) == 0) then
           tmp(1:4,1:3,1:ld%lat%volh,ip,0:1)=0
        endif
     enddo
     sf = dotprod(tmp, tmp, 2* SIZE_SC_FIELD)
  endif
  DEBUG2S("End: fermi_action_mtdagmt_hhbb")
end

