!===============================================================================
!
! cg_unprec.F90 - solves "matrix_mult * x = b" and returns number of iterations
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2008 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine cg_unprec(xe, xo, be, bo, id, iter, res) 
  use module_function_decl
  use module_check_solution
  use module_action
  use module_cg
  use module_vol
  use module_counter
  implicit none

  SPINCOL_OVERINDEXED,  intent(inout) :: xe, xo
  SPINCOL_OVERINDEXED,  intent(in)    :: be, bo
  integer,              intent(out)   :: iter
  integer,              intent(in)    :: id
  DOUBLE,               intent(in)    :: res

!!  SPINCOL_OVERINDEXED :: r, p, aap
  REAL, allocatable ::re(:),pe(:),aape(:)
  REAL, allocatable ::ro(:),po(:),aapo(:)
  REAL          :: ak, bk
  DOUBLE        :: rtr, rtrold, paap, bb, dtmpe, dtmp1e, dtmp2e, dtmpo, dtmp1o, dtmp2o
  integer       :: i, niter, ierr
  character(72) :: msg

!-------------------------------------------------------------------------------

  DEBUG2S("Start: cg_unprec")
  TIMING_START(timing_bin_cg)

  allocate(re(SIZE_COMPLEX*NDIRAC*NCOL*volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("cg: allocation failed re")
  endif
  allocate(pe(SIZE_COMPLEX*NDIRAC*NCOL*volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("cg: allocation failed pe")
  endif
  allocate(aape(SIZE_COMPLEX*NDIRAC*NCOL*volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("cg: allocation failed aape")
  endif
  allocate(ro(SIZE_COMPLEX*NDIRAC*NCOL*volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("cg: allocation failed ro")
  endif
  allocate(po(SIZE_COMPLEX*NDIRAC*NCOL*volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("cg: allocation failed po")
  endif
  allocate(aapo(SIZE_COMPLEX*NDIRAC*NCOL*volh_tot), STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("allocation failed", 1, ierr)
     call die("cg: allocation failed aapo")
  endif

  bb = 0.0_8
  !$omp parallel do reduction(+: bb) private(dtmpe, dtmpo)
  do i = 1, size_sc_field
     dtmpe = be(i)
     dtmpo = bo(i)
     bb = bb + dtmpe**2 + dtmpo**2
  enddo
  call global_sum_vec(1, bb)

  DEBUG2I("fermiID    ",  id)
  DEBUG2R("norm(b)=   ",  sqrt(bb)/2/action%mtilde(id)%kappa)
  DEBUG2R("rsd    =   ",  dble(res)**2*bb/(2*action%mtilde(id)%kappa)**2)


  call unprec_mdagm(re, ro, xe, xo, id)
  niter = 1
  rtrold = ZERO
  !$omp parallel do reduction(+: rtrold) private(dtmpe, dtmpo)
  do i = 1, size_sc_field
     re(i) = be(i) - re(i)
     ro(i) = bo(i) - ro(i)
     pe(i) = re(i)
     po(i) = ro(i)
     dtmpe = re(i)
     dtmpo = ro(i)
     rtrold = rtrold + dtmpe**2 + dtmpo**2
  enddo
  call global_sum_vec(1, rtrold)
  if (stop_condition(rtrold, res, bb)) goto 9999

  do niter = 1, cg_para%maxiter
     call unprec_mdagm(aape, aapo, pe, po, id)
     paap = ZERO
     !$omp parallel do reduction(+: paap) private(dtmp1e, dtmp1o, dtmp2e, dtmp2o)
     do i = 1, size_sc_field
        dtmp1e = pe(i)
        dtmp2e = aape(i)
        dtmp1o = po(i)
        dtmp2o = aapo(i)
        paap = paap + dtmp1e * dtmp2e + dtmp1o * dtmp2o
     enddo
     call global_sum_vec(1, paap)
     ak = rtrold / paap 

     rtr = 0.0_8
     !$omp parallel do reduction(+: rtr) private(dtmpe, dtmpo)
     do i = 1, size_sc_field
        xe(i) = xe(i) + ak * pe(i)
        re(i) = re(i) - ak * aape(i)
        dtmpe = re(i) 
        xo(i) = xo(i) + ak * po(i)
        ro(i) = ro(i) - ak * aapo(i)
        dtmpo = ro(i) 
        rtr = rtr + dtmpe**2 + dtmpo**2
     enddo 
     call global_sum_vec(1, rtr)

     if (stop_condition(rtr, res, bb)) goto 9999
     bk = rtr / rtrold
     rtrold = rtr

     call sc_xpby(pe, re, bk)  ! p = r + bk * p
     call sc_xpby(po, ro, bk)  ! p = r + bk * p
  enddo

  niter = niter - 1

  if (cg_para%log /= 2) then
     write(msg, *) "cg(): no convergence; rtr = ", rtr 
     call die(msg)
  endif

9999 continue

  call update_cg_stat(cg_stat, niter, 0)
  cg_iterations_total = cg_stat%niter_tot_run
  iter = niter

  TIMING_STOP(timing_bin_cg)

  if (cg_para%check /= 0) then  ! check solution

  endif

  deallocate(re, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("cg: deallocation failed re")
  endif
  deallocate(pe, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("cg: deallocation failed pe")
  endif
  deallocate(aape, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("cg: deallocation failed aape")
  endif
  deallocate(ro, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("cg: deallocation failed ro")
  endif
  deallocate(po, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("cg: deallocation failed po")
  endif
  deallocate(aapo, STAT = ierr)
  if (ierr /= 0) then
     call stderr2_int("deallocation failed", 1, ierr)
     call die("cg: deallocation failed aapo")
  endif

  DEBUG2S("End: cg_unprec")
end

!===============================================================================
