!===============================================================================
!
! gcrodr.F90 
!
! solves "matrix_mult * x = b" and returns number of iterations
!
! GCRO-DR Ref.:  Recycling krylov subspaces forsequences of linear systems  
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2008 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine gcrodr(x, b, id, iter, res) 
  use module_function_decl
  use module_check_solution
  use module_action
  use module_cg
  use module_vol
  use module_counter
  implicit none

  SPINCOL_OVERINDEXED,  intent(inout) :: x
  SPINCOL_OVERINDEXED,  intent(in)    :: b
  integer,              intent(out)   :: iter
  integer,              intent(in)    :: id
  DOUBLE,               intent(in)    :: res

  SPINCOL_OVERINDEXED :: r, aap
  REAL          :: ak, bk
  DOUBLE        :: rtr, paap, bb, dtmp, dtmp1, dtmp2
  integer       :: i, niter
  character(72) :: msg

!
! needed for checking solution
!
  DOUBLE        :: norm(3)
  integer, save :: headline_check_written = 0
  character(*), parameter :: key_chk = "%cgChk"
  character(*), parameter :: fmt_chk_h = "(1x, 2a, a6, 4a20)"
  character(*), parameter :: fmt_chk_b = "(1x, a7, i6, 3e20.10, i20)"


!-------------------------------------------------------------------------------

  TIMING_START(timing_bin_cg)

  bb = 0.0_8
  !$omp parallel do reduction(+: bb) private(dtmp)
  do i = 1, size_sc_field
     dtmp = b(i)
     bb = bb + dtmp**2
  enddo
  call global_sum_vec(1, bb)

  DEBUG2R("norm(b)=   ",  sqrt(bb)/2/action%mtilde(id)%kappa)
  DEBUG2R("rsd    =   ",  dble(res)**2*bb/(2*action%mtilde(id)%kappa)**2)

  call wdagw(r, x, id)
  rtrold = ZERO
  !$omp parallel do reduction(+: rtrold) private(dtmp)
  do i = 1, size_sc_field
     r(i) = b(i) - r(i)
     dtmp = r(i)
     rtr = rtr + dtmp**2
  enddo
  call global_sum_vec(1, rtr)
  if (stop_condition(rtr, res, bb)) goto 9999

!----------------------------------------

  if (ydefine) then
     call wdagw_mult_nsco(id, k, c, u)
     call qrdcmp_nsco(k, c, h)         ! R => h Q => new 
     call inv_upper_hessen(k, h)       ! h <= h^-1
     call mm_mult_nsco(k, u, h)        ! u <= u*h^-1
     call (k, c, x)    ! x <= x + U C^+ x
     call orthogonalize_sc(k, c, r)    ! r <= r - C C^+ r
  else
     call gmres_m(x, b, r, id, m, rtr)
!!   make Pk, line 14
     not yet
     call km_mult_nsco(k, m, u, v, p) !tilde{Y}_k = V_m P_k, line 15.

!!   QR dcmp of underbar{H}_m P_k, line 16
     call mm_mult(hp, uh ,p, m+1, k, m)
     call qrdcmp(hp, q, invr, m+1, k)

     call mm_mult_nsco(k, m+1, c, v, q)! C_k = V_{m+1} Q, line 17
     call mm_mult_nsco(k, u, invr)     !U_k = tilde{Y}_k R^{-1}, line 18
  endif

  do niter = 1, cg_para%maxiter
     if (stop_condition(rtr, res, bb)) goto 9999
     h=ZERO
     do i = 1, k
        v(i)%sco = u(i)%sco
        call normalize_sc(v(i)%sco, dummy)
        h(i,i) = ONE/dummy
     enddo
     v(k+1)%sco = r
     call normalize_sc(v(k+1)%sco, dummy)
     do j = k+1, m
        call wdagw(v(j+1)%sco, v(j)%sco, id)
        call orthogonalize_sc(k, c, (v(j+1)%sco) ! v <= (I - C C^+)Av
        do i = 1, j
           h(i,j) = sc_dot(v(j+1)%sco, v(i)%sco)
           h(i,j) = global_sum(h(i,j))
           call sc_axpy(v(j+1)%sco, v(i)%sco, -h(i,j))
        enddo
        call normalize_sc(v(j+1)%sco, h(j+1,j))
     enddo
     do j = k+1, m
        call wdagw(aap, v(j)%sco, id)
        do i = 1, k
           xx=sc_cdotc(c(i)%sco, aap)
           call global_sum_vec(2, xx)
           h(i,j) = xx
        enddo
     enddo

     htmp = h
     !!!!!!!!!!!!!!!!!!!!!!! koko kara
  enddo

  niter = niter - 1

  if (cg_para%log /= 2) then
     write(msg, *) "cg(): no convergence; rtr = ", rtr 
     call die(msg)
  endif

9999 continue

  call update_cg_stat(gcrodr_stat, niter, 0)
  iter = niter

  TIMING_STOP(timing_bin_cg)

  if (cg_para%check /= 0) then  ! check solution

     call check_solution(id, x, b, norm)

     if (headline_check_written == 0) then
        if (my_pe() == 0) then
           write(UREC, fmt_chk_h) &
               "T", key_chk, "traj", "rtr", "Rb", "Rx", "iterations"
        endif
        headline_check_written = 1
     endif
     
     if (my_pe() == 0) then
        write(UREC, fmt_chk_b) &
           key_chk, counter%traj, rtr, &
           norm(1)/norm(2), norm(1)/norm(3), iter
     endif

  endif

end

!-------------------------------------------------------------------------------
subroutine qrdcmp(input, q, invr, n, k) 
  implicit none

  COMPLEX, intent(in) :: input(n,k)
  COMPLEX, intent(out) :: q(n,k), invr(k,k)
  integer, intent(in)    :: n, k

end

!-------------------------------------------------------------------------------
subroutine mv_mult(out, in1, in2, c, r) 
  implicit none

  COMPLEX, intent(out) :: out(c)
  COMPLEX, intent(out) :: in1(c,r), in2(r)
  integer, intent(in)    :: c, r
  integer :: i, j, k

  !$omp parallel do
  do i = 1, c
     out(i) = in1(i,1) * in2(1)
  enddo
  do j = 2, r
  !$omp parallel do
     do i = 1, c
        out(i) = out(i) + in1(i,j) * in2(j)
     enddo
  enddo

end

!-------------------------------------------------------------------------------
subroutine mm_mult(out, in1, in2, c, r, m) 
  implicit none

  COMPLEX, intent(out) :: out(c, r)
  COMPLEX, intent(out) :: in1(c,m), in2(m,r)
  integer, intent(in)    :: c, r, m
  integer :: i, j, k

  do j = 1, r
     !$omp parallel do
     do i = 1, c
        out(i,j) = in1(i,1) * in2(1,j)
     enddo
  enddo

  do k = 2, m
     do j = 1, r
        !$omp parallel do
        do i = 1, c
           out(i,j) = out(i,j) + in1(i,k) * in2(k,j)
        enddo
     enddo
  enddo

end

!-------------------------------------------------------------------------------
subroutine qrdcmp_nsco(n, v, h) 
  use typedef
  use module_function_decl
  implicit none

  type(sco_field),intent(inout) :: v(n)
  REAL, intent(out) :: h(n,n)
  integer, intent(in) :: n
  integer :: i, j

  h=ZERO
  do j = 1, n
     do j = 1, j-1
        h(i,j) = sc_dot(v(i)%sco, v(j)%sco)
        h(i,j) = global_sum(h(i,j))
        call sc_axpy(v(i)%sco, v(j)%sco, -h(i,j))
     enddo
     call normalize_sc(v(i)%sco, h(j,j))
  enddo

end

!-------------------------------------------------------------------------------
subroutine wdagw_mult_nsco(id, n, out, in) 
  use typedef
  implicit none

  type(sco_field),intent(out) :: out(n)
  type(sco_field),intent(in) :: in(n)
  integer, intent(in) :: id, n
  integer :: i

  do i = 1, n
     call wdagw(out(i)%sco, in(i)%sco, id)
  enddo

end

!===============================================================================
