/*
!===============================================================================
!
! ildg.c - read/write configuration in ILDG lime format
!
! ildg_open_r() : open ILDG file for reading
! ildg_open_w() : open ILDG file for writing
! ildg_read()   : read one xy-slice
! ildg_write()  : write one xy-slice
! ildg_close()  : close ILDG file
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007 Hinnerk Stueben
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
*/
# include <sys/types.h>
# include <stdio.h>
# include <string.h>
# include "c_defs.h"
# include "c_util.h"
# include "cksum.h"
# include "lime.h"
# include "ildg.h"
# include "ildg_xml.h"

static FILE *fp;
static LimeReader *r;
static LimeWriter *w;
static LimeRecordHeader *h;

static char xmldoc[1024];
static int len_xmldoc = 1023;

/*----------------------------------------------------------------------------*/
void ILDG_OPEN_R(char *filename, 
                 int *Lx, int *Ly, int *Lz, int *Lt,
                 int *precision,
                 INT8 *bytes_total)
{
  /*  filename         (input)
   *  Lx, Ly, Lz, Lt   lattice size (output)
   *  precision        32|64 (output)
   *  bytes_total     bytes of configuration (output)
   */

  fp = fopen(filename, "r");
  if (fp == NULL) 
      die("could not open %s for reading", filename);

  r = limeCreateReader(fp);

  ildg_find_record("ildg-format");
  ildg_read_string("ildg-format", xmldoc, len_xmldoc);

  ildg_xml_check_value(xmldoc, "<version>", "1.0");
  ildg_xml_check_value(xmldoc, "<field>", "su3gauge");
  ildg_xml_get_int(xmldoc, "<precision>", precision);
  ildg_xml_get_int(xmldoc, "<lx>", Lx);
  ildg_xml_get_int(xmldoc, "<ly>", Ly);
  ildg_xml_get_int(xmldoc, "<lz>", Lz);
  ildg_xml_get_int(xmldoc, "<lt>", Lt);

  ildg_find_record("ildg-binary");

  *bytes_total = limeReaderBytes(r);

  CKSUM_INIT();
}

/*----------------------------------------------------------------------------*/
void ILDG_OPEN_W(char *filename, 
                 int *Lx, int *Ly, int *Lz, int *Lt,
                 int *precision,
                 INT8 *bytes_total)
{
  /*  filename         (input)
   *  Lx, Ly, Lz, Lt   lattice size (input)
   *  precision        32|64 (input)
   *  bytes_total     bytes of configuration (input)
   */

  n_uint64_t nbytes;

  fp = fopen(filename, "w");
  if (fp == NULL)
      die("could not open %s for writing", filename);

  w = limeCreateWriter(fp);

  sprintf(xmldoc, 
          "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
          "<ildgFormat>\n"
          "  <version>1.0</version>\n"
          "  <field>su3gauge</field>\n"
          "  <precision>%d</precision>\n"
          "  <lx>%d</lx> <ly>%d</ly> <lz>%d</lz> <lt>%d</lt>\n"
          "</ildgFormat>\n",
          *precision, *Lx, *Ly, *Lz, *Lt);

  nbytes = strlen(xmldoc);

  h = limeCreateHeader(1, 0, "ildg-format", nbytes);

  limeWriteRecordHeader(h, w);
  limeDestroyHeader(h);
  limeWriteRecordData(xmldoc, &nbytes, w);

  nbytes = *bytes_total;

  h = limeCreateHeader(0, 1, "ildg-binary-data", nbytes);
  limeWriteRecordHeader(h, w);
  limeDestroyHeader(h);

  CKSUM_INIT();
}

/*----------------------------------------------------------------------------*/
void ILDG_READ(void *buffer, INT8 *bytes)
{
  /*  buffer        (input)
   *  bytes         (input)
   */

  int status;
  n_uint64_t nbytes = *bytes;

  status = limeReaderReadData(buffer, &nbytes, r);

  CKSUM_ADD(buffer, bytes);
}

/*----------------------------------------------------------------------------*/
void ILDG_WRITE(void *buffer, INT8 *bytes)
{
  /*  buffer        (output)
   *  bytes         (input)
   */
  int status;
  n_uint64_t nbytes = *bytes;

  status = limeWriteRecordData(buffer, &nbytes, w);
  
  CKSUM_ADD(buffer, bytes);
}

/*----------------------------------------------------------------------------*/
void ILDG_CLOSE_R(INT8 *cksum, INT8 *bytes_total, char *lfn, int *len)
{
  /*  cksum         CRC check sum (output)
   *  bytes_total   (output) (from cksum())
   *  lfn           logical file name (output)
   *  len           input: max. length of "lfn"
   *                output: actual length of "lfn"
   */

  n_uint64_t nbytes;
  int status;

  fseek(fp, 0, SEEK_SET);
  do {
      status = limeReaderNextRecord(r);
      if (status == LIME_EOF)
          break;
  } while (strncmp(limeReaderType(r), "ildg-data-lfn", 13) != 0);


  if (status == LIME_EOF)
      strcpy(lfn, "UNDEFINED");
  else
      ildg_read_string("ildg-data-lfn", lfn, *len);

  *len = strlen(lfn);

  limeDestroyReader(r);
  fclose(fp);

  CKSUM_GET(cksum, bytes_total);
}

/*----------------------------------------------------------------------------*/
void ILDG_CLOSE_W(INT8 *cksum, INT8 *bytes_total, char *lfn)
{
  /*  cksum         CRC check sum (output)
   *  bytes_total   (output) (from cksum())
   *  lfn           logical file name (input)
   */

  n_uint64_t nbytes = strlen(lfn);

  h = limeCreateHeader(1, 1, "ildg-data-lfn", nbytes);
  limeWriteRecordHeader(h, w);
  limeDestroyHeader(h);
  limeWriteRecordData(lfn, &nbytes, w);

  limeDestroyWriter(w);
  fclose(fp);
  
  CKSUM_GET(cksum, bytes_total);
}

/*----------------------------------------------------------------------------*/
void ildg_find_record(char *record)
{
  /*  record         (input)
   */

  int status;

  fseek(fp, 0, SEEK_SET);
  do {
      status = limeReaderNextRecord(r);
      if (status == LIME_EOF)
          die("unexpected end-of-file");
  } while (strncmp(limeReaderType(r), record, (int) strlen(record)) != 0);
}

/*----------------------------------------------------------------------------*/
void ildg_read_string(char *record, char *s, int len)
{
  /*  record         (input)
   *  s              string read (output)
   *  len            max. length of "s"
   */

  int status;
  n_uint64_t nbytes;

  nbytes = limeReaderBytes(r);
  if (nbytes > len)
    die("string (length=%d) too small for reading record %s", len, record);

  status = limeReaderReadData(s, &nbytes, r);
  if (status != LIME_SUCCESS)
      die("failed to read record %s (status=%d)", record, status);
  s[nbytes] = '\0';
}

/*============================================================================*/
