!===============================================================================
!
! ildg_conf.F90 - read/write configurations in ILDG format
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007-2008 Hinnerk Stueben
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
module module_ildg_conf

  character(*), parameter :: READ = "read"
  character(*), parameter :: WRITE = "write"
  character(*), parameter :: LFN_KEY = "bqcd-restart"

  character(*), parameter :: fmt_restart = "(a, '.', i3.3, '.', 2a)"
  character(*), parameter :: fmt_conf    = "(a, '.', i3.3, '.', i5.5, '.', 2a)"
 
  character, parameter    :: ascii_null = char(0)

  character(256)          :: lfn
  integer                 :: lx, ly, lz, lt, precision
  INT8                    :: words, bytes, cksum
end

!-------------------------------------------------------------------------------
subroutine conf_read_ildg(restart, para, conf)

  use typedef_para
  use module_function_decl
  use module_input
  use module_p_interface
  use module_vol
  use module_ildg_conf
  implicit none

  logical                               :: restart
  type(type_para)                       :: para
  type(hmc_conf), dimension(MAX_TEMPER) :: conf

  P_GAUGE_FIELD_ILDG, save              :: u_ildg
  FILENAME                              :: filename
  integer                               :: len_lfn
  INT8                                  :: bytes_read

  call begin(UREC, "ILDGread")

  ALLOCATE_G_FIELD_ILDG(u_ildg)

  call ildg_lime_file(filename, READ, restart)
  call ildg_open_r(filename, lx, ly, lz, lt, precision, bytes)
  call ildg_check()
  call ildg_set_sizes()
  
  TIMING_START(timing_bin_u_read_ildg)
  call ildg_io(READ, u_ildg)
  TIMING_STOP(timing_bin_u_read_ildg)

  len_lfn = len(lfn) - 1
  call ildg_close_r(cksum, bytes_read, lfn, len_lfn)

  if (my_pe() == 0) then
     call ildg_check2(restart)
     call ildg_rec(READ, filename)

     ASSERT(bytes == bytes_read)
  endif

  call ildg_seq(READ, conf(1)%u, u_ildg)
  call xbound_g_field(conf(1)%u)
  conf(1)%former = 1

  call end(UREC, "ILDGread")
end

!-------------------------------------------------------------------------------
subroutine conf_write_ildg(restart, para, conf)

  use typedef_para
  use module_function_decl
  use module_input
  use module_p_interface
  use module_vol
  use module_ildg_conf
  implicit none

  logical                               :: restart
  type(type_para)                       :: para
  type(hmc_conf), dimension(MAX_TEMPER) :: conf

  P_GAUGE_FIELD_ILDG, save              :: u_ildg
  FILENAME                              :: filename
  INT8                                  :: cksum0, bytes_written
  integer                               :: i
  REAL, external                        :: sg
  REAL                                  :: plaquette

  call begin(UREC, "ILDGwrite")

  ALLOCATE_G_FIELD_ILDG(u_ildg)

  plaquette  = ONE - (sg(conf(1)%u) / (SIX * volume))

  call ildg_lime_file(filename, WRITE, restart)

  lx = para%l(1)
  ly = para%l(2)
  lz = para%l(3)
  lt = para%l(4)
  
  if (restart) then
     precision = input%ildg_precision_restart
  else
     precision = input%ildg_precision
  endif

  call ildg_set_sizes()

  call ildg_check()
  call ildg_seq(WRITE, conf(1)%u, u_ildg)
  
  if (my_pe() == 0) then
     call ildg_open_w(filename, lx, ly, lz, lt, precision, bytes)
  endif

  TIMING_START(timing_bin_u_write_ildg)
  call ildg_io(WRITE, u_ildg)
  TIMING_STOP(timing_bin_u_write_ildg)

  if (my_pe() == 0) then
     call cksum_get(cksum, bytes_written)
     call ildg_set_lfn(filename, restart)
     call ildg_close_w(cksum0, bytes_written, lfn)
     call ildg_rec(WRITE, filename)

     if (.not. restart) then
        i = index(lfn, ascii_null)
        ASSERT(i > 0)
 call ildg_meta("ensemble", para, conf, lfn(1:i-1), plaquette, precision, cksum)
 call ildg_meta("conf", para, conf, lfn(1:i-1), plaquette, precision, cksum)
     endif

     ASSERT(cksum0 == cksum)
     ASSERT(bytes == bytes_written)
  endif

  call end(UREC, "ILDGwrite")
end

!-------------------------------------------------------------------------------
subroutine ildg_set_sizes()

  use module_ildg_conf
  use module_vol
  implicit none
  integer :: bytes_per_word

  select case (precision)
     case(32)
        bytes_per_word = 4
     case(64)
        bytes_per_word = 8
     case default
        call die("ildg_set_sizes(): unknown precision")
  end select

  words = SIZE_COMPLEX * NCOL * NCOL * DIM * volume

  bytes = words
  bytes = bytes * bytes_per_word        ! 64 bit integer arithmetic

end  

!-------------------------------------------------------------------------------
subroutine ildg_check()

  use module_ildg_conf
  use module_input
  implicit none
  INT8 :: bytes0

  CHECK(input%lattice(1) == lx)
  CHECK(input%lattice(2) == ly)
  CHECK(input%lattice(3) == lz)
  CHECK(input%lattice(4) == lt)
  CHECK(precision == 32 .or. precision == 64)

  bytes0 = bytes
  call ildg_set_sizes()
  CHECK(bytes == bytes0)

end

!-------------------------------------------------------------------------------
subroutine ildg_check2(restart)

  use module_counter
  use module_ildg_conf
  implicit none
  logical, intent(in) :: restart

  character :: a, b
  integer   :: run, job, traj, i
  INT8      :: cksum0, bytes0

  if (index(lfn, LFN_KEY) > 0) then
     i = index(lfn, ascii_null)
     ASSERT(i > 0)
     read(lfn(1:i-1), *) a, b, cksum0, bytes0, run, job, traj
     CHECK(cksum == cksum0)
     if (restart) then
        CHECK(run == counter%run)
        CHECK(job + 1 == counter%job)
        CHECK(traj == counter%traj)
     endif
  endif

end

!-------------------------------------------------------------------------------
subroutine ildg_filename(fname, action, restart, extension)

  use module_bqcd
  use module_counter
  use module_ildg_conf
  use module_input
  implicit none
  character(*), intent(out) :: fname
  character(*), intent(in)  :: action
  logical, intent(in)       :: restart
  character(*), intent(in)  :: extension
  integer :: i

  if (restart) then
     write(fname, fmt_restart) &
           trim(prog_name), counter%run, trim(extension), ascii_null
  else
     if (action == READ) then
        if (input%start_ildg_file == "") then
           call die("ildg_filename(): start_ildg_file not set")
        else
           fname = input%start_ildg_file
           i = index(fname, " ")
           CHECK(i > 0)
           CHECK(i <= len(fname))
           fname(i:i) = ascii_null
        endif
     elseif (action == WRITE) then
        write(fname, fmt_conf) &
              trim(input%ildg_filename_prefix), &
              counter%run, counter%traj, trim(extension), ascii_null
     else
        call die("ildg_filename(): unknown action")
     endif
  endif
end

!-------------------------------------------------------------------------------
subroutine ildg_lime_file(fname, action, restart)

  use module_input
  implicit none
  character(*), intent(out) :: fname
  character(*), intent(in)  :: action
  logical, intent(in)       :: restart

  call ildg_filename(fname, action, restart, input%ildg_filename_extension)
end

!-------------------------------------------------------------------------------
subroutine ildg_set_lfn(fname, restart)

  use module_counter
  use module_input
  use module_ildg_conf
  implicit none

  character(*), intent(in)  :: fname
  logical, intent(in)       :: restart
  integer                   :: i

  if (restart) then

     ! save counters and checksum in the lime file instead of lfn

     i = index(fname, ascii_null)

     write(lfn, *) LFN_KEY, " ", &
                   fname(1:i-1), &
                   cksum, bytes, &
                   counter%run, counter%job, counter%traj, &
                   ascii_null
  else
     call ildg_expand_input_lfn_mc()
     lfn = trim(input%ildg_data_lfn_path) // "/" // trim(fname) // ascii_null
  endif

end

!-------------------------------------------------------------------------------
subroutine ildg_seq(action, u, u_out)

  use module_ildg_conf
  use module_lattice
  use module_vol
  implicit none
  character(*), intent(in)      :: action
  GAUGE_FIELD, intent(in)       :: u
  GAUGE_FIELD_ILDG, intent(out) :: u_out

  if (precision == 32) call ildg_seq_r4(action, u, u_out)
  if (precision == 64) call ildg_seq_r8(action, u, u_out)

end

!-------------------------------------------------------------------------------
subroutine ildg_io(action, u)

  use module_ildg_conf
  use module_lattice_io
  use module_vol
  implicit none
  character(*), intent(in)        :: action
  GAUGE_FIELD_ILDG, intent(inout) :: u

  if (precision == 32) call ildg_io_r4(action, u, words, bytes)
  if (precision == 64) call ildg_io_r8(action, u, words, bytes)
end

!-------------------------------------------------------------------------------
subroutine ildg_rec(action, fname)

  use module_ildg_conf
  implicit none
  character(*), intent(in) :: action
  character(*), intent(in) :: fname

  integer :: i, j

  i = index(fname, ascii_null); ASSERT(i > 0)
  j = index(lfn, ascii_null);   ASSERT(j > 0)

  write(UREC, "(x,4a)") "ildg-", action, " file ", fname(1:i-1)
  write(UREC, *)        "ildg-", action, " precision ", precision
  write(UREC, *)        "ildg-", action, " bytes ", bytes
  write(UREC, *)        "ildg-", action, " cksum ", cksum
  write(UREC, "(x,4a)") "ildg-", action, " lfn ", lfn(1:j-1)

end

!===============================================================================
