!===============================================================================
!
! integrator.F90
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2008 Yoshifumi Nakamura 
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
module module_integrator

  integer, save :: n_time_scale, tot_pq_loop
  integer, dimension(:), pointer, save  :: scale, m_scale, jscale
  REAL, dimension(:), pointer, save  :: step, step_p
  REAL, dimension(:,:), pointer, save  :: step_q, step_ppq
  character(len=6), pointer, save :: integrator(:)

  integer,save :: itraj
  integer,save :: iscale
  REAL, save :: len

end

!-------------------------------------------------------------------------------
subroutine integrate(p, para, conf, sf1, sf2, sf3)
  use typedef_hmc
  use module_integrator
  use module_vol
  use module_sf
  implicit none

  GENERATOR_FIELD, intent(inout) :: p
  type(hmc_para),  intent(in)    :: para
  type(hmc_conf),  intent(inout) :: conf
  REAL,            intent(out)   :: sf1, sf2, sf3
  integer :: i, j
  logical, external :: zero_step_q

  DEBUG2S("Start MD step")
  TIMING_START(timing_bin_integrator)
  if (associated(sf))sf = ZERO

  DEBUG2S("Start init_integrator")
  call init_integrator(para%traj_length, para%tau)
  DEBUG2S("End init_integrator")

  do i = 0, tot_pq_loop - 1
     call hmc_integrator_q(p, para, conf, step_p(i))
     call hmc_integrator_p(p, para, conf, i, 0)
  enddo

  call hmc_integrator_q(p, para, conf, step_p(tot_pq_loop))
  if (zero_step_q(tot_pq_loop)) then
     call calc_fermi_action(.false.)
  else
     call hmc_integrator_p(p, para, conf, tot_pq_loop, 1)
  endif

  if (associated(sf)) then
     call hmc_sum_sf3(sf1,sf2,sf3)
  else
     sf1=0; sf2=0; sf3=0
  endif
  TIMING_STOP(timing_bin_integrator)
  DEBUG2S("End MD step")
end


!-------------------------------------------------------------------------------
subroutine init_integrator(traj_length, tau)
  use module_integrator
  implicit none

  REAL, intent(in) :: traj_length, tau
  integer,save :: count = 0
  integer :: i

  if (count /= 0 .and. tau > 0 ) return

  len = traj_length
  if (tau < 0) len = - len

!!  tot_pq_loop = scale(n_time_scale) * 2**(n_time_scale)
!!  tot_pq_loop = scale(n_time_scale) * 2**1 * 3**2
!!  tot_pq_loop = scale(n_time_scale) * 1**(n_time_scale)

  tot_pq_loop = scale(n_time_scale)
  do i = 1, n_time_scale
     select case (integrator(i))
        case ("LPFSTS"); tot_pq_loop = tot_pq_loop
        case ("LPFTST"); tot_pq_loop = tot_pq_loop * 2
        case ("2MNSTS"); tot_pq_loop = tot_pq_loop * 2
        case ("2MNTST"); tot_pq_loop = tot_pq_loop * 3
        case ("4MN4FP"); tot_pq_loop = tot_pq_loop * 5
        case ("4MN5FV"); tot_pq_loop = tot_pq_loop * 5
        case default   ; call die("subintegrator: unknown integrator" &
                         // integrator(iscale)  )
     end select
  enddo


  allocate(step_p(0:tot_pq_loop))
  allocate(step_q(0:tot_pq_loop,n_time_scale))

  step_p = ZERO
  step_q = ZERO
  itraj = 0
  iscale = 0
  call subintegrator(ONE)
  count = count + 1

end

!-------------------------------------------------------------------------------
subroutine subintegrator(fac)
  use module_integrator
  implicit none
  REAL, intent(in) :: fac

  select case (integrator(iscale+1))
     case ("LPFSTS"); call hmc_lpfsts_integrator(fac)
     case ("LPFTST"); call hmc_lpftst_integrator(fac)
     case ("2MNSTS"); call hmc_2mnsts_integrator(fac)
     case ("2MNTST"); call hmc_2mntst_integrator(fac)
     case ("4MN4FP"); call hmc_4mn4fp_integrator(fac)
     case ("4MN5FV"); call hmc_4mn5fv_integrator(fac)
     case default   ; call die("subintegrator: unknown integrator" &
                      // integrator(iscale)  )
  end select

end

!-------------------------------------------------------------------------------
logical function zero_step_q(i)
  use module_integrator
  implicit none
  integer, intent(in) :: i
  integer :: j
  REAL :: max_step

  max_step = ZERO

  do j = 1, n_time_scale
     max_step = max(max_step, step_q(i,j))
  enddo

  zero_step_q = (max_step == ZERO)

end

!===============================================================================

