!===============================================================================
!
! traces.F90
!
! calculates: Tr(inv(M))          psibar psi (pbp)
!             Tr(gamma5 inv(M))   psibar gamma5 psi (p5p)
!             Tr(inv(M+ M))       pion norm  (pinorm)
!
! traces of a matrix A are calculated with a stochastic estimator:
!
!     Tr(A) = eta+ A eta  (eta: Gaussian noise)
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2007 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
subroutine traces(para, conf, traj, i_ensemble1, i_ensemble2)
  use module_field
  use module_action
  use typedef_hmc
  use module_input
  use module_function_decl
  use module_p_interface
  use module_vol
  use module_cg
  implicit none

  integer,        intent(in)  :: traj, i_ensemble1, i_ensemble2
  type(hmc_para), intent(in)  :: para
  type(hmc_conf), intent(in)  :: conf

  P_SPINCOL_FIELD, save       :: eta_e, eta_o, zeta_e, zeta_o

  character(len=*), parameter :: key_tr = "%tr"
  integer,        save        :: count = 0
  integer                     :: n_sc_field, size_of_trace
  integer                     :: cg_ncall, cg_niter_max, cg_niter_tot
  integer                     :: fid, mid, cid
  REAL                        :: pinorm
  REAL                        :: re_pbp, im_pbp, re_p5p, im_p5p
  COMPLEX                     :: pbp, p5p
  REAL                        :: res(5)

  fid = max(input%hmc_mpf_dd_mass,input%hmc_mpf_mass)
  fid = max(input%hmc_mpf_hh_mass,fid)
  fid = max(input%hmc_mpf_un_mass,fid)
  fid = max(input%hmc_mpf_rhmc_l,fid)
  fid = max(input%hmc_mpf_dd_rhmc_l,fid)
  fid = max(input%hmc_mpf_hh_rhmc_l,fid)

  if (fid <= 0 .or. fid > 5 ) call die("traces: unexpected fid")

  mid = action%fermi(fid)%mid1

  ALLOCATE_SC_FIELD(eta_e)
  ALLOCATE_SC_FIELD(eta_o)
  ALLOCATE_SC_FIELD(zeta_e)
  ALLOCATE_SC_FIELD(zeta_o)

  count = count + 1
  n_sc_field = NDIRAC * NCOL * volh
  size_of_trace = NDIRAC * NCOL * volume

  call init_cg_stat(cg_stat)
  call init_imp_all(.true.)

  call make_source(eta_e, eta_o, RANDOM_SOURCE)
!  call make_source(eta_e, eta_o, FULL_SOURCE)

  call solve(mid, zeta_e, zeta_o, eta_e, eta_o)  ! zeta = inv(M) eta  ! original
  call check_solution(mid, zeta_e, zeta_o, eta_e, eta_o)


  pbp = sc_cdotc(eta_e, zeta_e) + sc_cdotc(eta_o, zeta_o)
  pinorm = sc_dot(zeta_e, zeta_e) + sc_dot(zeta_o, zeta_o)
  call gamma5(zeta_e, volh)  ! zeta = gamma5 inv(M) eta
  call gamma5(zeta_o, volh)
  p5p = sc_cdotc(eta_e, zeta_e) + sc_cdotc(eta_o, zeta_o)
  res(1) = Re(pbp) ;  res(2) = Im(pbp)
  res(3) = Re(p5p) ;  res(4) = Im(p5p)
  res(5) = pinorm
  call global_sum_vec(5, res)
  re_pbp = res(1) / size_of_trace
  im_pbp = res(2) / size_of_trace
  re_p5p = res(3) / size_of_trace
  im_p5p = res(4) / size_of_trace
  pinorm = res(5) / size_of_trace
  call get_cg_stat(cg_stat, cg_ncall, cg_niter_max, cg_niter_tot)

  if (my_pe() == 0) then
     if (count == 1) write(UREC, 400) &
         "T", key_tr, "traj", "e", "f", &
         "Re(pbp)", "Im(pbp)", "Re(p5p)", "-Im(p5p)", "PionNorm", "CGiter"

     write(UREC, 410) key_tr, traj, i_ensemble1, i_ensemble2, &
         re_pbp, im_pbp, re_p5p, -im_p5p, pinorm, cg_niter_max
  endif

!!---------------------------------------------
!!--------------------------------------- for s
!  call init_cg_stat(cg_stat)
!  fid = input%hmc_mpf_mass + input%hmc_mpf_rhmc_s
!  mid = action%fermi(fid)%mid1
!  call solve(mid, zeta_e, zeta_o, eta_e, eta_o)  ! zeta = inv(M) eta
!  call check_solution(mid, zeta_e, zeta_o, eta_e, eta_o)
!  pbp = sc_cdotc(eta_e, zeta_e) + sc_cdotc(eta_o, zeta_o)
!  pinorm = sc_dot(zeta_e, zeta_e) + sc_dot(zeta_o, zeta_o)
!  call gamma5(zeta_e, volh)  ! zeta = gamma5 inv(M) eta
!  call gamma5(zeta_o, volh)
!  p5p = sc_cdotc(eta_e, zeta_e) + sc_cdotc(eta_o, zeta_o)
!  res(1) = Re(pbp)
!  res(2) = Im(pbp)
!  res(3) = Re(p5p)
!  res(4) = Im(p5p)
!  res(5) = pinorm
!  call global_sum_vec(5, res)
!  re_pbp = res(1) / size_of_trace
!  im_pbp = res(2) / size_of_trace
!  re_p5p = res(3) / size_of_trace
!  im_p5p = res(4) / size_of_trace
!  pinorm = res(5) / size_of_trace
!  call get_cg_stat(cg_stat, cg_ncall, cg_niter_max, cg_niter_tot)
!  if (my_pe() == 0) then
!     write(UREC, 410) key_tr, traj, i_ensemble1, i_ensemble2, &
!         re_pbp, im_pbp, re_p5p, -im_p5p, pinorm, cg_niter_max
!  endif

400 format (1x, 2a, a6, 2a3, 5a20,    a10)
410 format (1x, a4, i6, 2i3, 5g20.10, i10)

end

!-------------------------------------------------------------------------------
subroutine solve(mid, out_e, out_o, in_e, in_o)  ! solves:  M out = in
  use module_input
  use module_dd
  use module_vol
  implicit none
  integer, intent(in) :: mid
  SPINCOL_FIELD,  intent(out) :: out_e, out_o
  SPINCOL_FIELD,  intent(in)  :: in_e, in_o

  if (ddable) then
     call test_mmul_mdag(mid)
     if (input%fullsolver=="eo") then
        call eo_solve(mid, out_e, out_o, in_e, in_o)
     else
        !!call dd_solve(mid, out_e, out_o, in_e, in_o)
        call dd_solve2(mid, out_e, out_o, in_e, in_o, CG_MC)
     endif
  else
     call eo_solve(mid, out_e, out_o, in_e, in_o)
  endif
end

!-------------------------------------------------------------------------------
!
!   (Tee Meo) =  (1   Meo Tee^-1)*(\tilde{M}ee 0  )*(1            0)
!   (Moe Too)    (0      1      ) (0           Too) (Too^-1 Moe   1)
!                 ~~~~~~~~~~~~~~                     ~~~~~~~~~~~~~~
!                       U                                  L
!   \tilde{M}ee = Tee - Meo Too^-1 Moe
!
!-------------------------------------------------------------------------------
subroutine eo_solve(mid, out_e, out_o, in_e, in_o)  ! solves:  M out = in
  use module_field
  use module_action
  use module_vol
  use module_p_interface
  use module_function_decl
  implicit none

  integer, intent(in) :: mid
  SPINCOL_FIELD,  intent(out) :: out_e, out_o
  SPINCOL_FIELD,  intent(in)  :: in_e, in_o
  P_SPINCOL_FIELD, save :: tmp
  
  REAL     :: a, b, res
  integer  :: iterations, cid

  DEBUG2S("Start: solve_eo")
  ALLOCATE_SC_FIELD(tmp)

  a = ONE / (ONE + action%mtilde(mid)%h**2)
  b = action%mtilde(mid)%kappa / (ONE + action%mtilde(mid)%h**2)
  cid = action%mtilde(mid)%cid

! in : xi
! out : phi

  ! (Moo)^{-1} xi_o
  if (action%mtilde(mid)%cswkappa /= ZERO) then
     tmp = in_o
     call clover_mult_ao(clover(cid)%i(1,1,so), tmp, volh)
     out_o = tmp
  elseif (action%mtilde(mid)%h /= ZERO) then
     call h_mult_c(out_o, -action%mtilde(mid)%h, in_o, volh)
  else
     out_o = in_o
  endif

  ! xi_e =  xi_e - Meo Moo^-1 xi_o
  call d(se, so, out_e, out_o, gauge(2)%u)
  call sc_xpby(out_e, in_e, b)

!  phi_e = phi_e = (\tilde{M}^{\dagger}\tilde{M})^{-1}  \tilde{M}^{\dagger} xi_e
#ifdef OMTDTD
  if (action%mtilde(mid)%cswkappa /= ZERO) then
     call clover_mult_ao(clover(cid)%i(1,1,se), out_e, volh)
  else
     if (action%mtilde(mid)%h /= ONE) &
     call h_mult_b( real(action%mtilde(mid)%h,kind=RKIND), out_e, volh)
  endif
#endif
  call mtil_dag(tmp, out_e, mid)
  call cg_outer(out_e, tmp, mid, 0, CG_MC)

!  Moe phi_e
  call d(so, se, out_o, out_e, gauge(2)%u)

!  xi_o - Moe phi_e
  call sc_axpby(out_o, in_o, b, a)!(inout, in, b, a)  ! inout = b * inout + a * in

!  phi_o = Moo^{-1} ( xi_o - Moe phi_e )
  if (action%mtilde(mid)%cswkappa /= ZERO) then
     call clover_mult_ao(clover(cid)%i(1,1,so), out_o, volh)
  else
     if (action%mtilde(mid)%h /= ZERO) call h_mult_b(-action%mtilde(mid)%h, out_o, volh)
  endif

  DEBUG2S("End: solve_eo")
end

!-------------------------------------------------------------------------------
subroutine gamma5(x, volh)

  implicit none
  COMPLEX, dimension (NDIRAC, *) :: x
  integer                        :: volh 

  integer :: i
  COMPLEX :: x1, x2, x3, x4

  !$omp parallel do private(x1, x2, x3, x4)
  do i = 1, NCOL * volh
     x1 = x(1, i)
     x2 = x(2, i)
     x3 = x(3, i)
     x4 = x(4, i)
#ifdef GAMMAC
     x(1, i) = x1
     x(2, i) = x2
     x(3, i) =-x3
     x(4, i) =-x4
#else     
     x(1, i) = x3
     x(2, i) = x4
     x(3, i) = x1
     x(4, i) = x2
#endif
  enddo
end

!-------------------------------------------------------------------------------
subroutine check_solution(mid, zeta_e, zeta_o, eta_e, eta_o)
  use module_field
  use module_action
  use module_vol
  use module_p_interface
  use module_function_decl
  implicit none

  integer, intent(in) :: mid  
  SPINCOL_FIELD,  intent(in) :: zeta_e, zeta_o, eta_e, eta_o
  P_SPINCOL_FIELD, save :: tmp1, tmp2
  REAL     :: kappa, residual1, residual2, bb
  integer  :: cid

  ALLOCATE_SC_FIELD(tmp1)
  ALLOCATE_SC_FIELD(tmp2)

  cid = action%mtilde(mid)%cid
  kappa = action%mtilde(mid)%kappa

  bb=sc_norm2(eta_e) + sc_norm2(eta_o) ; bb=global_sum(bb)


  ! (1) Mee zeta_e + Meo zeta_o = eta_e
  ! (2) Moe zeta_e + Moo zeta_o = eta_o

  ! check (1)
  if (cid /=0) then
     call clover_mult_a(tmp1, clover(cid)%a(1,1,se), zeta_e, volh)
  else
     tmp1 = zeta_e
  endif
  call d(se, so, tmp2, zeta_o, gauge(2)%u)
  call sc_xpby(tmp2, tmp1, - kappa)
  tmp2 = tmp2 - eta_e
  residual1 =  sc_norm2(tmp2);  residual1 =  global_sum(residual1)

  ! check (2)
  if (cid /= 0) then
     call clover_mult_a(tmp1, clover(cid)%a(1,1,so), zeta_o, volh)
  else
     tmp1=zeta_o
  endif
  call d(so, se, tmp2, zeta_e, gauge(2)%u)
  call sc_xpby(tmp2, tmp1, - kappa)
  tmp2 = tmp2 - eta_o
  residual2 =  sc_norm2(tmp2);  residual2 =  global_sum(residual2)
  if (my_pe()==0)write(0,*) "residual of solution |b-Dx|/|b|: ", &
  sqrt(residual1+residual2)/sqrt(bb)

end

!-------------------------------------------------------------------------------
subroutine make_propagator()
  use module_field
  use module_action
  use module_input
  use module_function_decl
  use module_p_interface
  use module_vol
  use module_cg
  implicit none

  P_SPINCOL_FIELD, save       :: in_e, in_o, out_e, out_o
  SECONDS                     :: time0, sekunden, time
  integer                     :: fid, mid, cid, iters, mu, col
  DOUBLE :: res

  fid = max(input%hmc_mpf_dd_mass,input%hmc_mpf_mass)
  fid = max(input%hmc_mpf_hh_mass,fid)
  fid = max(input%hmc_mpf_un_mass,fid)

  if (fid <= 0 .or. fid > 5 ) call die("traces: unexpected fid")
  mid = action%fermi(fid)%mid1

  ALLOCATE_SC_FIELD(in_e)
  ALLOCATE_SC_FIELD(in_o)
  ALLOCATE_SC_FIELD(out_e)
  ALLOCATE_SC_FIELD(out_o)

  call init_cg_stat(cg_stat)
  call init_imp_all(.true.)

  if (my_pe()==0) write(0,*) "WALL SOURCE"
  call make_source_wall(1,1,0,in_e,in_o)
  call solve(mid, out_e, out_o, in_e, in_o)  ! solves:  M out = in

  if (my_pe()==0) then
     write(0,*)"solver summary:CG"
     write(0,*)"first 12 components"
     do col=1,3
     do mu=1,4
     write(0,*)mu,col,out_e(mu,col,1)
     enddo
     enddo
  endif

end

!===============================================================================
